"""Module to configure OpalKelly XEM7310 FPGA to communicate with peripherals.

Use Registers for peripherals' internal registers and Endpoints for Opal Kelly Endpoints on the FPGA. 

Abe Stroschein, ajstroschein@stthomas.edu

Lucas Koerner, koer2434@stthomas.edu
"""

import pandas as pd
import os
import sys
import copy
import yaml
from .utils import gen_mask, str_bitfile_version
from warnings import warn

home_dir = os.path.join(os.path.expanduser('~'), '.pyripherals')
config_path = os.path.join(home_dir, 'config.yaml')
if os.path.exists(config_path):
    with open(config_path, 'r') as file:
        configs = yaml.safe_load(file)

    # TODO: make sure this path works for Windows, Mac, and Linux: https://docs.opalkelly.com/fpsdk/frontpanel-api/programming-languages/
    if sys.platform == 'win32':
        sys.path.append(os.path.join(configs['frontpanel_path'], 'API/Python/3.7/x64')) # Add path to ok.py to PATH for import
        dll_dir = os.path.join(configs['frontpanel_path'], 'API/lib/x64')
        os.add_dll_directory(dll_dir)   # Make DLL (okFrontPanel.dll) available
        os.environ.setdefault('PATH', '')
        os.environ['PATH'] += os.pathsep + dll_dir
    elif sys.platform == 'darwin':
        sys.path.append(os.path.join(configs['frontpanel_path'], 'API/Python3/')) # Add _ok.so to PATH for import
    try:
        import ok
    except ImportError:
        warn(message=f"Could not find FrontPanel files. Is \n\tfrontpanel_path: {configs['frontpanel_path']} from {config_path} the correct path?", category=ImportWarning)
else:
    warn(message=f'No config.yaml file found at {config_path}. Use pyripherals.utils.create_yaml to create one.', category=UserWarning)
    from .utils import DEFAULT_CONFIGS
    configs = DEFAULT_CONFIGS


class Register:
    """Class for internal registers on a device.

    Attributes
    ----------
    address : int
        Address location of the register.
    default : int
        Default value of the register.
    bit_index_high : int
        Index of the MSB in the register.
    bit_index_low : int
        Index of the LSB in the register.
    bit_width : int
        Width of the register in bits.
    """

    def __init__(self, address, default, bit_index_high, bit_index_low, bit_width):
        self.address = address
        self.default = default
        self.bit_index_high = bit_index_high
        self.bit_index_low = bit_index_low
        self.bit_width = bit_width

    def __str__(self):
        return f'{hex(self.address)}[{self.bit_index_low}:{self.bit_index_high}]'

    def __eq__(self, other):
        if type(self) is type(other):
            return self.__dict__ == other.__dict__
        else:
            return False

    @staticmethod
    def get_chip_registers(sheet, workbook_path=configs['registers_path']):
        """Return a dictionary of Registers from a page in an Excel spreadsheet."""

        reg_dict = {}
        sheet_data = pd.read_excel(workbook_path, sheet)
        for row in range(len(sheet_data)):
            row_data = sheet_data.iloc[row]
            reg_dict[row_data['Name']] = Register(
                address=int(row_data['Hex Address'], 16),
                default=int(row_data['Default Value'], 16),
                bit_width=int(row_data['Bit Width']),
                bit_index_high=(None if row_data['Bit Index (High)'] == 'None' else int(
                    row_data['Bit Index (High)'])),  # Bit Index of None means the register takes up the whole endpoint
                bit_index_low=(None if row_data['Bit Index (Low)'] == 'None' else int(row_data['Bit Index (Low)'])))
        return reg_dict


class Endpoint:
    """Class for Opal Kelly endpoints on the FPGA.

    Attributes
    ----------
    endpoints_from_defines : dict
        Class attribute. Dictionary of each group of endpoints paired with inner dictionaries
        of endpoint names to Endpoint objects, starts empty.
    I2CDAQ_level_shifted : dict
        Class attribute. Dictionary of Endpoints for the level shifted I2CDAQ bus.
    I2CDAQ_QW : dict
        Class attribute. Dictionary of Endpoints for QW 3.3V I2CDAQ bus.
    address : int
        Address location of the Endpoint.
    bit_index_low : int
        Index of the LSB of the Endpoint.
    bit_index_high : int
        Index of the MSB of the Endpoint.
    bit_width : int
        Width of the Endpoint in bits. This number is used when using advance_endpoints if gen_bit is True.
    gen_bit : bool
        Whether to increment the bits when incrementing the Endpoint.
    gen_address : bool
        Whether to increment the address when incrementing the Endpoint.
    addr_step : int
        How much to add to the address when using advance_endpoints if gen_address is True.
    """

    MAX_WIDTH = configs['endpoint_max_width']  # Maximum bit width of an Endpoint. Used to wrap Endpoints to the next address when incrementing
    endpoints_from_defines = dict()
    I2CDAQ_level_shifted = dict()
    I2CDAQ_QW = dict()

    def __init__(self, address, bit_index_low, bit_width, gen_bit, gen_address, addr_step=1):
        self.address = address
        self.bit_index_low = bit_index_low
        # Endpoints that are only containing addresses will be generated from ep_defines.v with bit_index_low = None
        if bit_index_low == None:
            self.bit_index_high = None
        else:
            self.bit_index_high = bit_index_low + bit_width
        self.bit_width = bit_width
        self.gen_bit = gen_bit
        self.gen_address = gen_address
        self.addr_step = addr_step

    def __str__(self):
        str_rep = '0x{:0x}[{}:{}]'.format(
            self.address, self.bit_index_low, self.bit_index_high)
        return str_rep

    def __eq__(self, other):
        if type(self) is type(other):
            return self.__dict__ == other.__dict__
        else:
            return False

    @staticmethod
    def update_endpoints_from_defines(ep_defines_path=configs['ep_defines_path']):
        """Store and return a dictionary of Endpoints for each chip in ep_defines.v.

        Returns -1 if there is a naming collision in ep_defines.v
        """

        # Leave as None if there is no ep_defines.v
        if ep_defines_path is None:
            warn(f'No ep_defines_path set in your {config_path} file')
            return dict()
        elif not os.path.exists(ep_defines_path):
            raise FileNotFoundError(f'ep_defines_path = {ep_defines_path} not found')

        # Get all lines
        with open(ep_defines_path, 'r') as file:
            lines = file.readlines()

        # Get all the endpoints from the lines
        for line in lines:
            # Check if the line defines an endpoint
            pieces = line.split(' ')
            # Ex. line = "`define AD7961_PIPE_OUT_GEN_ADDR 8'hA1 // address=TEST_ADDRESS bit_width=32"
            #   pieces = ["`define", "AD7961_PIPE_OUT_GEN_ADDR", "8'hA1", "//", "address=TEST_ADDRESS", "bit_width=32"]
            if pieces[0] != '`define':
                # Line does not define an endpoint, skip
                continue

            # Grab "addr_step" and remove from pieces to leave later steps the same
            shortened_pieces = [x[:len('addr_step')] for x in pieces]
            if 'addr_step' in shortened_pieces:
                addr_step_text = pieces.pop(shortened_pieces.index('addr_step'))
                try:
                    addr_step = addr_step_text[len('addr_step='):]
                    addr_step = int(addr_step)
                except ValueError as e:
                    raise ValueError(f'addr_step not assigned to int in line {lines.index(line) + 1} of "{ep_defines_path}". Got addr_step={addr_step} instead.')
            else:
                # No addr_step assigned, default to 1
                addr_step = 1

            # Extract data from definition
            # Class name
            class_name_end = pieces[1].find('_')
            if class_name_end == -1:
                # .find() returns -1 if not found, without an underscore we cannot
                # tell where the class and endpoint names are separated
                print('FAIL: no endpoint name found')
                print(line)
                continue
            class_name = pieces[1][:class_name_end]

            # Whether to generate bits
            remaining_name = pieces[1][class_name_end + 1:]
            previous_len = len(remaining_name)
            remaining_name = remaining_name.replace('_GEN_BIT', '')
            # If the remaining name is shorter it is because GEN_BIT was replaced
            gen_bit = len(remaining_name) < previous_len

            # Whether to generate address
            previous_len = len(remaining_name)
            remaining_name = remaining_name.replace('_GEN_ADDR', '')
            # If the remaining name is shorter it is because GEN_ADDR was replaced
            gen_address = len(remaining_name) < previous_len

            # Endpoint name
            ep_name = remaining_name
            if ep_name == 'NUM_OUTGOING_EPS':
                # This is not an endpoint and does not match the
                # format of the others so we skip it
                continue

            # Address, bit, and bit_width
            if "8'h" in pieces[2]:
                # Definition holds an address, take that value
                address = int(pieces[2][3:], base=16)
                bit = None
                bit_width = int(pieces[4].split('=')[1])
            else:
                # Definition holds a bit, take address from comment
                comment_address = pieces[4].split('=')[1]
                if '0x' in comment_address:
                    # Address comment has a hex value
                    address = int(comment_address[2:], 16)
                else:
                    # Address comment has the name of another endpoint so store
                    # the interfaces.py name of that endpoint so we can look it
                    # up going through all lines
                    address_name = pieces[4].split('=')[1]
                    address_name_no_gen = address_name.split(
                        '_GEN', maxsplit=1)[0]
                    address = address_name_no_gen
                bit = int(pieces[2])
                bit_width = int(pieces[5].split('=')[1])

            endpoint = Endpoint(address=address, bit_index_low=bit,
                                bit_width=bit_width, gen_bit=gen_bit, gen_address=gen_address, addr_step=addr_step)

            # Put defined endpoint in endpoints_from_defines dictionary
            if Endpoint.endpoints_from_defines.get(class_name) is None:
                # Class doesn't exist yet in the dictionary
                Endpoint.endpoints_from_defines[class_name] = {
                    ep_name: endpoint}
            else:
                # Class already exists in the dictionary
                Endpoint.endpoints_from_defines[class_name][ep_name] = endpoint

        # Go through endpoints_from_defines and find hex addresses for those with endpoint
        # name references instead
        for group_name in Endpoint.endpoints_from_defines:
            group = Endpoint.endpoints_from_defines[group_name]
            for endpoint_name in group:
                endpoint = group[endpoint_name]
                if type(endpoint.address) == str:
                    # Address is a name
                    class_name, ep_name = endpoint.address.split(
                        '_', maxsplit=1)
                    referenced_group = Endpoint.endpoints_from_defines.get(
                        class_name)
                    if referenced_group is None:
                        print(
                            f'{group_name}[{endpoint_name}]: Referenced group "{class_name}" not found.')
                        continue
                    referenced_endpoint = referenced_group.get(ep_name)
                    if referenced_endpoint is None:
                        print(
                            f'{group_name}[{endpoint_name}]: Referenced endpoint "{class_name}_{ep_name}" not found.')
                        continue
                    endpoint.address = referenced_endpoint.address

        # At this point the dictionary should be built
        # Check for naming collisions (2+ names sharing same address or bit within address)
        # Collect all top level endpoints
        top_level_eps = [
            x for x in Endpoint.endpoints_from_defines.values() if type(x) == Endpoint]
        # Collect all endpoints in dictionaries
        lower_level_eps = []
        for eps in [d.values() for d in Endpoint.endpoints_from_defines.values() if type(d) == dict]:
            lower_level_eps += list(eps)
        none_to_neg_1 = {None: -1}
        # Make tuples of (address, bit) from each list, then concatenate the lists
        # The none_to_neg_1.get() translates None values of bit_index_low to -1 for sorting later
        list_eps = [(ep.address, none_to_neg_1.get(
            ep.bit_index_low, ep.bit_index_low)) for ep in top_level_eps]
        list_len = len(list_eps)
        set_len = len(set(list_eps))  # A set removes duplicates

        if list_len != set_len:
            # There may be duplicates. May not, because different groups can have endpoints of the same name.
            print('Checking for naming collisions in ep_defines.v ...')

            # Search through to find duplicates
            # Copy the list to keep order in the original
            sorted_list_eps = list(list_eps)
            sorted_list_eps.sort()  # Put duplicates next to one another in new list
            top_level_names = [x for x in Endpoint.endpoints_from_defines.keys() if type(
                Endpoint.endpoints_from_defines[x]) == Endpoint]
            lower_level_names = []
            for sub_dicts in [x for x in Endpoint.endpoints_from_defines.values() if type(x) == dict]:
                lower_level_names += list(sub_dicts.keys())
            list_names = top_level_names + lower_level_names

            collision = False  # Keep track of whether there was a collision for return value
            for ep_index in range(len(sorted_list_eps) - 1):
                ep = sorted_list_eps[ep_index]
                next_ep = sorted_list_eps[ep_index + 1]
                name = list_names[list_eps.index(sorted_list_eps[ep_index])]
                next_name = list_names[list_eps.index(
                    sorted_list_eps[ep_index])]
                if (ep == next_ep) and (name != next_name):
                    # Need different names because otherwise they are from different groups and do not actually conflict

                    # The name part of this uses a list of names created in the same order as the original list_eps (list_names)
                    # then finds the index of the current endpoint in list_eps and uses that to find the corresponding
                    # name in list_names
                    print(
                        f'Collision found at address={ep[0]} bit={ep[1]}: {name} with {next_name}')
                    collision = True
            if collision:
                return -1
            else:
                print('No collisions found.')

        # If the list and set match length, no duplicates
        return Endpoint.endpoints_from_defines

    @staticmethod
    def get_chip_endpoints(chip_name):
        """Return a copy of the dictionary of Endpoints for a specific chip or group."""

        if Endpoint.endpoints_from_defines == dict():
            Endpoint.update_endpoints_from_defines()

        # copy.deepcopy() is important here because it makes a copy of the
        # dictionary so that when we increment it for multiple instantiations
        # of the chip, each previously instantiated chip will not have its
        # endpoints affected by the increment, only future instantiations.
        # Using deepcopy() ensures that any dictionaries inside the dictionary
        # we copy also get copied, not left as references.
        return copy.deepcopy(Endpoint.endpoints_from_defines.get(chip_name))

    @staticmethod
    def excel_to_defines(excel_path, defines_path, sheet=0):
        """Convert an Excel spreadsheet of endpoint definitions to Verilog.

        Parameters
        ----------
        excel_path : str
            The path to the Excel spreadsheet to convert.
        defines_path : str
            The path to the Verilog file to create.
        sheet : int or str
            Optional. The int index of the sheet to read from the Excel
            spreadsheet, or the str sheet name.

        Returns
        -------
        str : the text written to the Verilog file.
        """

        sheet_data = pd.read_excel(excel_path, sheet)
        text = '\n'.join(sheet_data['Generated Line'])
        with open(defines_path, 'w') as file:
            file.write(text)
        return text

    @staticmethod
    def advance_endpoints(endpoints_dict, advance_num=1):
        """
        Advances Endpoints in a dict in place by advance_num.
        
        Checks each Endpoint's gen_bit and gen_address attributes to see
        whether to increment the bit or the address or both. The Endpoint's
        bit_width and addr_step attributes determine how much to increment the
        bits or address by. If the Endpoint's bits would exceed
        Endpoint.MAX_WIDTH (determined from config.yaml), then the bits wrap
        around to start at bit zero on the next address determined by adding
        the Endpoint's addr_step attribute.

        Example usage:
            endpoints=Endpoint.advance_endpoints(Endpoint.get_chip_endpoints('I2CDAQ'),1)

        Parameters
        ----------
        endpoints_dict : dict of Endpoints
            The dict of Endpoints to advance by advance_num.
        advance_num : int
            How much to advance the Endpoints by.

        Returns
        -------
        dict : the same dict of Endpoints given in endpoints_dict, now advanced
        """

        for key in endpoints_dict:
            endpoint = endpoints_dict[key]
            if endpoint.gen_bit:
                endpoint.bit_index_low += (endpoint.bit_width * advance_num)
                if endpoint.bit_index_low > Endpoint.MAX_WIDTH:
                    # Endpoint does not fit, wrap to next address
                    endpoint.address += endpoint.addr_step
                    endpoint.bit_index_low %= Endpoint.MAX_WIDTH
                endpoint.bit_index_high = endpoint.bit_index_low + endpoint.bit_width
                if endpoint.bit_index_high > Endpoint.MAX_WIDTH:
                    # Endpoint split across two addresses -> move to next address, start at bit 0
                    endpoint.address += endpoint.addr_step
                    endpoint.bit_index_low = 0
                    endpoint.bit_index_high = endpoint.bit_index_low + endpoint.bit_width
            if endpoint.gen_address:
                endpoint.address += advance_num * endpoint.addr_step
        return endpoints_dict

# Class for the FPGA itself. Handles FPGA configuration, setting wire values,
# and other FPGA specific functions.


class FPGA:
    """Class for the Opal Kelly FPGA itself.
    Derived from OpalKelly Python examples.

    Attributes
    ----------
    bitfile : str
        Path to the bitfile to load on the FPGA.
    xem : ok.okCFrontPanel
        Opal Kelly API connection to the FPGA.
    device_info : ok.okTDeviceInfo
        General information about the FPGA.
    """


    def __init__(self, bitfile='default', endpoints=None, debug=False):
        if bitfile == 'default':
            # Use bitfile from config.yaml fpga_bitfile_path
            self.bitfile = configs['fpga_bitfile_path']
        else:
            self.bitfile = bitfile

        if endpoints is None:
            self.endpoints = Endpoint.get_chip_endpoints('GP')
        else:
            self.endpoints = endpoints

        self.debug = debug
        self.bitfile_version = None


    def init_device(self):
        """Initialize the FPGA for use and print device information.

        Connect to the FPGA and load the bitfile. Return False on any errors.
        Only run this once or the FPGA connection will fail.
        """

        # Open the first device we find.
        self.xem = ok.okCFrontPanel()
        if (self.xem.NoError != self.xem.OpenBySerial("")):
            print("A device could not be opened.  Is one connected?")
            return(False)

        # Get some general information about the device.
        self.device_info = ok.okTDeviceInfo()
        if (self.xem.NoError != self.xem.GetDeviceInfo(self.device_info)):
            print("Unable to retrieve device information.")
            return(False)
        print("         Product: " + self.device_info.productName)
        print("Firmware version: %d.%d" %
              (self.device_info.deviceMajorVersion, self.device_info.deviceMinorVersion))
        print("   Serial Number: %s" % self.device_info.serialNumber)
        print("       Device ID: %s" % self.device_info.deviceID)
        print("       USB Speed: %d" % self.device_info.usbSpeed)

        self.xem.LoadDefaultPLLConfiguration()

        try:
            self.bitfile_version = self.read_wire(self.endpoints['BITFILE_VERSION'].address)
        except KeyError:
            print('No Endpoint BITFILE_VERSION in Endpoints read from', configs['ep_defines_path'])
            # Default to version 1
            self.bitfile_version = 1
        if self.bitfile_version < 0:
            # An error occurred in the read
            print('Error reading BITFILE_VERSION endpoint:', self.bitfile_version)
            self.bitfile_version = 1    # 00.00.01
        print('Bitfile Version:', str_bitfile_version(self.bitfile_version))

        # Download the configuration file.
        if self.bitfile is not None:
            if (self.xem.NoError != self.xem.ConfigureFPGA(self.bitfile)):
                print("FPGA configuration failed.")
                return(False)
            else:
                print('Loaded bit-file: {}'.format(self.bitfile))
        else:
            print('Skipped bit-file update')

        # Check for FrontPanel support in the FPGA configuration.
        if (False == self.xem.IsFrontPanelEnabled()):
            print("FrontPanel support is not available.")
            return(False)

        print("FrontPanel support is available.")
        return self

    def read_pipe_out(self, addr, data_len=1024):
        """Return the filled buffer and error code after reading an OK PipeOut.
            data_len is length in bytes (must be multiple of 16)
            returns: bytearray; error code
        """
        buf = bytearray(data_len)
        e = self.xem.ReadFromPipeOut(addr, buf)
        # print('read_pipe_out:', addr, buf)

        if (e < 0):
            print('Error code {}'.format(e))
        return buf, e

    def set_wire(self, address, value, mask=0xFFFFFFFF):
        """Return the error code after setting an OK WireIn value."""
        if self.debug:
            print(
                f'set_wire(address={hex(address)}, value={hex(value)}, mask={hex(mask)})')
        error_code = self.xem.SetWireInValue(address, value, mask)
        self.xem.UpdateWireIns()
        return error_code

    def read_wire(self, address):
        """Return the read data after reading an OK WireOut."""

        self.xem.UpdateWireOuts()
        return self.xem.GetWireOutValue(address)

    def set_endpoint(self, ep_bit):
        """Set all bits in an Endpoint high."""

        mask = gen_mask(
            list(range(ep_bit.bit_index_low, ep_bit.bit_index_high + 1)))
        if self.debug:
            print(
                f'set_endpoint(address={hex(ep_bit.address)}, value={hex(mask)}, mask={hex(mask)})')
        self.xem.SetWireInValue(ep_bit.address, mask, mask)  # set
        self.xem.UpdateWireIns()

    def clear_endpoint(self, ep_bit):
        """Set all bits in an Endpoint low."""

        mask = gen_mask(
            list(range(ep_bit.bit_index_low, ep_bit.bit_index_high + 1)))
        if self.debug:
            print(
                f'clear_endpoint(address={hex(ep_bit.address)}, value={hex(0)}, mask={hex(mask)})')
        self.xem.SetWireInValue(ep_bit.address, 0x0000, mask)  # clear
        self.xem.UpdateWireIns()

    def toggle_low(self, ep_bit):
        """Toggle all bits in an Endpoint low then back to high."""

        mask = gen_mask(
            list(range(ep_bit.bit_index_low, ep_bit.bit_index_high + 1)))
        self.xem.SetWireInValue(
            ep_bit.address, 0x0000, mask)  # toggle low
        self.xem.UpdateWireIns()
        self.xem.SetWireInValue(ep_bit.address, mask, mask)   # back high
        self.xem.UpdateWireIns()

    def toggle_high(self, ep_bit):
        """Toggle all bits in an Endpoint high then back to low."""

        mask = gen_mask(
            list(range(ep_bit.bit_index_low, ep_bit.bit_index_high + 1)))
        self.xem.SetWireInValue(ep_bit.address, mask, mask)  # toggle high
        self.xem.UpdateWireIns()
        self.xem.SetWireInValue(ep_bit.address, 0x0000, mask)   # back low
        self.xem.UpdateWireIns()

    def send_trig(self, ep_bit):
        """Return the error code after activating an OK TriggerIn Endpoint.

        Expects a single bit, not yet implement for multiple bits and will only
        activate the LSB if the Endpoint containts multiple bits.
        """

        # print(f'send_trig(address={hex(ep_bit.address)},bit={ep_bit.bit_index_low})')
        return self.xem.ActivateTriggerIn(ep_bit.address, ep_bit.bit_index_low)

    def read_trig(self, ep_bit):
        """Read an OK TriggerOut Endpoint.
        
        Parameters
        ----------
        ep_bit : Endpoint
            The endpoint containing the bit_index_low and address of the
            TriggerOut to read.

        Returns
        -------
        bool
            Whether the TriggerOut has been triggered.
        """

        self.xem.UpdateTriggerOuts()
        return self.xem.IsTriggered(ep_bit.address, (1 << ep_bit.bit_index_low))

    def read_ep(self, ep_bit):
        """Return the error code after reading an OK WireOut Endpoint."""
        self.xem.UpdateWireOuts()
        read_out = self.xem.GetWireOutValue(ep_bit.address)
        return read_out

    def set_wire_bit(self, address, bit):
        """Set a single bit to 1 in a OpalKelly wire in."""

        if self.debug:
            print(
                f'set_wire_bit(address={hex(address)},value={hex(1 << bit)},mask={hex(1 << bit)})')
        return self.set_wire(address, value=1 << bit, mask=1 << bit)

    def clear_wire_bit(self, address, bit):
        """Clear a single bit to 0 in a OpalKelly wire in."""
        if self.debug:
            print(
                f'clear_wire_bit(address={hex(address)},value={hex(1 << bit)},mask={hex(1 << bit)})')

        return self.set_wire(address, value=0, mask=1 << bit)

    def set_ep_simultaneous(self, address, bit_list, val_list):
        """ set multiple values to the wire of a single endpoint"""
        value = 0
        mask = 0
        for bit, val in zip(bit_list, val_list):
            value = value | (val << bit)
            mask = mask | (1 << bit)
        if self.debug:
            print(
                f'Simultaneous wire write (address={hex(address)},value={hex(value)},mask={hex(mask)})')

        return self.set_wire(address, value=value, mask=mask)


    def read_wire_bit(self, address, bit):
        """Read a single bit in a OpalKelly wire in."""
        value = self.read_wire(address)
        return (value & (1 << bit)) >> bit


# TODO: should there be a 'device' or similar class that all controllers are subclasses of?
def disp_device(dev, reg=True):
    """Display endpoints and registers for a chip.

    Parameters
    ----------
    dev : I2CController or SPIController or SPIFifoDriven or AD7961
        Chip instance to display information for.
    reg : bool
        Whether to display registers.
    """

    print('Displaying endpoints')
    for k in dev.endpoints:
        print(k + ': ' + str(dev.endpoints[k]))

    print('-'*40)
    if reg:
        print('Displaying registers')
        try:
            for r in dev.registers:
                print(r + ': ' + str(dev.registers[r]))
        except:
            pass



from ..core import Register
from .I2CController import I2CController


class TCA9555(I2CController):
    """Class for the I/O Expander TCA9555.

    Subclass of the I2CController class. Attributes and methods below are
    differences in this class from I2CController only.

    Attributes
    ----------
    ADDRESS_HEADER : int
        7-bit device address with R/W bit space and 4 MSBs filled in specific
        to this chip. The rest is left as 0 to be filled in later.
    registers : dict
        Name-Register pairs for the internal registers of the TCA9555 chip.
    addr_pins : int
        3 LSBs of the 7-bit device address formed alongside the address header
        used to differentiate between different instances of the TCA9555 chip.
    """

    ADDRESS_HEADER = 0b0100_0000
    registers = Register.get_chip_registers('TCA9555')

    def configure_pins(self, data):
        """Configure the chip's pins as inputs (1's) or outputs (0's)."""

        # Inputs = 1
        # Outputs = 0
        dev_addr = self.ADDRESS_HEADER + (self.addr_pins << 1)
        self.i2c_write_long(
            dev_addr, [TCA9555.registers['CONFIG'].address], 2, data)

    def write(self, data, register_name='OUTPUT', mask=0xffff):
        """Write 2 bytes of data to the pins."""

        dev_addr = self.ADDRESS_HEADER | (self.addr_pins << 1)

        # Compare with current data to mask unchanged values
        if mask == 0xffff:
            new_data = data
        else:
            read_out = self.read()
            current_data = (read_out[0] << 8) | read_out[1]
            if current_data == None:
                print('Read for masking FAILED')
                return False
            new_data = (data & mask) | (current_data & ~mask)

        # Turn data into a list of bytes for i2c_write method
        list_data = [new_data // (16**2), new_data % (16**2)]
        self.i2c_write_long(
            dev_addr, [TCA9555.registers[register_name].address], 2, list_data)

    def read(self, register_name='INPUT'):
        """Read 2 bytes of data from the pins."""

        dev_addr = self.ADDRESS_HEADER | (self.addr_pins << 1) | 0b1
        return self.i2c_read_long(dev_addr, [TCA9555.registers[register_name].address], 2)



from ..core import Endpoint, Register
import copy


class SPIController:
    """Class for controllers on the FPGA using SPI protocol.

    Attributes
    ----------
    WB_SET_ADDRESS : int
        Set address command for the Wishbone.
    WB_WRITE : int
        Write command for the Wishbone.
    WB_READ : int
        Read command for the Wishbone.
    WB_CLK_FREQ : int
        Frequency of the clock the Wishbone runs on.
    fpga : FPGA
        FPGA instance this controller uses to communicate.
    endpoints : dict
        Endpoints on the FPGA this controller uses to communicate.
    master_config : int
        Value of the CTRL register in the Wishbone.
    registers : dict
        Name-Register pairs for the internal registers of the Wishbone.
    """

    WB_SET_ADDRESS = 0x80000000  # These 3 are from the SPI core manual
    WB_WRITE = 0x40000000
    WB_READ = 0x00000000
    # ACK=0x200000000
    WB_CLK_FREQ = 200  # clk_sys = 200 MHz in the top_level_module.v comments

    registers = Register.get_chip_registers('SPI')

    def __init__(self, fpga, endpoints, master_config=registers.get('CTRL').default):
        self.fpga = fpga
        self.master_config = master_config
        self.endpoints = endpoints

    @classmethod
    def create_chips(cls, fpga, number_of_chips, endpoints=None, master_config=None):
        """Instantiate a number of new chips.

        The number must be an integer greater than zero. The endpoints between
        each instance will be incremented. If the endpoints argument is left
        as None, then we will use copies of the endpoints_from_defines
        dictionary for the endpoints for each instance, and update that
        original dictionary when we increment the endpoints. This way, the
        endpoints there are ready for another instantiation if needed.
        """

        if type(number_of_chips) is not int or number_of_chips <= 0:
            print('number_of_chips must be an integer greater than 0')
            return False

        chips = []
        if master_config is None:
            # Use class default for master_config
            for i in range(number_of_chips):
                chips.append(cls(fpga=fpga, endpoints=endpoints))
                # Use deepcopy to keep the endpoints for different instances separate
                endpoints = copy.deepcopy(chips[-1].endpoints)
                Endpoint.advance_endpoints(endpoints)
        else:
            for i in range(number_of_chips):
                chips.append(cls(fpga=fpga, endpoints=copy.deepcopy(
                    endpoints), master_config=master_config))
                # Use deepcopy to keep the endpoints for different instances separate
                endpoints = copy.deepcopy(chips[-1].endpoints)
                Endpoint.advance_endpoints(endpoints)

        if endpoints is None:
            # Increment shared endpoints dictionary
            # We need to get the shared endpoints in endpoints_from_defines to
            # increment and we only have the endpoints given to us in the
            # argument.
            shared_full_eps = Endpoint.endpoints_from_defines
            shared_chip_eps = shared_full_eps[
                list(shared_full_eps.keys())[
                    list(shared_full_eps.values()).index(chips[0].endpoints)
                ]
            ]
            Endpoint.advance_endpoints(shared_chip_eps, number_of_chips)
        else:
            # Increment custom dictionary
            Endpoint.advance_endpoints(endpoints, number_of_chips)

        return chips

    def wb_send_cmd(self, command):
        """Send a command to the Wishbone.

        Sent in 32 bits through the WbSignal_converter Verilog module tos
        reformat to 34 bits.
        """

        # print(f'wb_send_cmd: {hex(self.endpoints["WB_IN"].address)}, {hex(command)}')
        self.fpga.set_wire(
            self.endpoints['WB_IN'].address, command, 0xffffffff)
        # DEBUG: temporary for logging DAC80508 test
        # print(f'ActivateTriggerIn(address={hex(self.parameters["WB_CONVERT"].address)}, index={hex(self.parameters["WB_CONVERT"].bit_index_high)}')

        # allows for optional debug prints
        self.fpga.send_trig(self.endpoints['WB_CONVERT'])
        #self.fpga.xem.ActivateTriggerIn(
        #    self.endpoints['WB_CONVERT'].address, self.endpoints['WB_CONVERT'].bit_index_low)  # High and low bit indexes are the same for the trigger because it is 1 bit wide

    def wb_is_ack(self):
        """Check if the Wishbone sent back a write acknowledge."""

        pass
    #     response = self.fpga.read_wire(self.endpoints['OUT'].address)
    #     return response == SPIController.ACK

    def wb_set_address(self, address):
        """Set the address of the register we interact with in the Wishbone."""

        self.wb_send_cmd(
            SPIController.WB_SET_ADDRESS | (address << 2) | 0x1)  # TODO: comment here what the | 0x1 is for when I figure it out
        return self.wb_is_ack()  # TODO: test if acknowledge actually comes back

    def wb_write(self, data):
        """Write up to 30 bytes of data to the currently selected register."""

        self.wb_send_cmd(SPIController.WB_WRITE | data)
        return self.wb_is_ack()  # TODO: test if acknowledge actually comes back

    def wb_read(self):
        """Return the data stored in the currently selected register."""

        self.wb_send_cmd(SPIController.WB_READ)
        return self.fpga.read_wire(self.endpoints['OUT'].address)

    def wb_go(self):
        """Initiate a SPI transmission.

        Initiated by setting the GO_BSY bit of the CTRL register to  1.
        """

        ack = self.wb_set_address(SPIController.registers['CTRL'].address)
        # TODO: fix ack
        # if ack:
        #     if self.debug:
        #         print(f'Set address to "CTRL": SUCCESS')
        # else:
        #     if self.debug:
        #         print(f'Set address to "CTRL": FAIL')
        #     return False

        # Use or so we do not overwrite the CTRL configuration, just the GO_BSY bit
        busy = self.master_config | (
            0x1 << SPIController.registers['GO_BSY'].bit_index_low)
        # add mask or something to not reset entire CTRL register
        ack = self.wb_write(busy)
        # TODO: fix ack
        # if ack:
        #     if self.debug:
        #         print(f'Set GO_BSY bit: SUCCESS')
        # else:
        #     if self.debug:
        #         print(f'Set GO_BSY bit: FAIL')
        #     return False

        return True

    def select_slave(self, slave_address):
        """Selecting a slave device.

        Set the correspoinding bit in the Slave Select register to 1."""

        # Set address to SS register
        self.wb_set_address(SPIController.registers['SS'].address)

        # Write slave address
        # Don't need to mask the write because we only select 1 slave device at a time
        self.wb_write(slave_address)

    def set_divider(self, divider):
        """Set the value of the Wishbone's DIVIDER register."""

        self.wb_set_address(SPIController.registers['DIVIDER'].address)
        self.wb_write(divider)

    def set_frequency(self, frequency):
        """Set the frequency of SPI SCLK.

        Set the DIVIDER register of the Wishbone with the corresponding divider
        for the target frequency. Frequency must be given in MHz.
        """

        divider = max(
            round((SPIController.WB_CLK_FREQ) / (frequency * 2)), 1) - 1
        # Make sure divider does not exceed 32 bits
        divider = min(divider, 0xffff_ffff)
        # print('Set frequency. Divider value = {}'.format(divider))
        self.set_divider(divider)
        return divider

    def write(self, data, register=0):
        """Write data on the SPI lines.

        Register should be either 0, 1, 2, or 3.
        """

        # Set address to Tx register
        ack = self.wb_set_address(
            SPIController.registers['Tx' + str(register)].address)
        # TODO: fix ack
        # if ack:
        #     if self.debug:
        #         print(f'Set address to "Tx{register}": SUCCESS')
        # else:
        #     if self.debug:
        #         print(f'Set address to "Tx{register}": FAIL')
        #     return False

        # Write data
        ack = self.wb_write(data)
        # TODO: fix ack
        # if ack:
        #     if self.debug:
        #         print(f'Data write to "Tx{register}": SUCCESS')
        # else:
        #     if self.debug:
        #         print(f'Data write to "Tx{register}": FAIL')
        #     return False

        # Set GO_BSY bit
        return self.wb_go()

    def read(self, register=0):
        """Return data from the selected data receive register.

        Register should be 0, 1, 2, or 3.
        """

        # Set address to the Rx register Rx0, Rx1, Rx2, or Rx3
        ack = self.wb_set_address(
            SPIController.registers['Rx' + str(register)].address)

        # TODO: fix ack
        # if ack:
        #     if self.debug:
        #         print(f'Set address to "Rx{register}": SUCCESS')
        # else:
        #     if self.debug:
        #         print(f'Set address to "Rx{register}": FAIL')
        #     return False

        # Read data
        return self.wb_read()

    def configure_master_bin(self, data):
        """Set the value of the Wishbone's CTRL register directly."""

        self.wb_set_address(SPIController.registers['CTRL'].address)
        self.wb_write(data)
        self.master_config = data

    def configure_master(self, ASS=0, IE=0, LSB=0, Tx_NEG=0, Rx_NEG=0, CHAR_LEN=0):
        """Set the Wishbone's CTRL register using several parameters.

        Parameters
        ----------
        ASS : bool or int
            Automatic Slave Select
        IE : bool or int
            Set interrupt output active after a transfer is finished
        LSB : bool or int
            Send LSB first
        Tx_NEG : bool or int
            Change output signal on falling edge of SCLK
        Rx_NEG : bool or int
            Latch input signal on falling edge of SCLK
        CHAR_LEN : bool or int
            Number of bits transmitted per transfer (up to 64).
        """

        params = {'ASS': ASS, 'IE': IE, 'LSB': LSB,
                  'Tx_NEG': Tx_NEG, 'Rx_NEG': Rx_NEG, 'CHAR_LEN': CHAR_LEN}

        configuration = SPIController.registers['CTRL'].default
        for param in params:
            configuration |= (
                params[param] << SPIController.registers[param].bit_index_low)
        self.configure_master_bin(configuration)

    def get_master_configuration(self):
        """Return the current configuration of CTRL register.

        Includes ASS, IO, LSB, Tx_NEG, Rx_NEG, CHAR_LEN.
        """

        self.wb_set_address(SPIController.registers['CTRL'].address)
        config_data = self.wb_read()
        #TODO: convert to text data to show individual settings
        # return config
        return config_data

    def reset_master(self):
        """Reset the Wishbone Master and SPI Core."""

        self.fpga.xem.ActivateTriggerIn(
            self.endpoints['MASTER_RESET'].address,
            self.endpoints['MASTER_RESET'].bit_index_low)

    def set_host_mode(self):
        """ Configure SPI controller to be host driven."""

        self.fpga.set_wire_bit(self.endpoints['HOST_FPGA'].address,
                               self.endpoints['HOST_FPGA'].bit_index_low)

    def set_fpga_mode(self):
        """ Configure SPI controller to be host driven."""

        self.fpga.clear_wire_bit(self.endpoints['HOST_FPGA'].address,
                                 self.endpoints['HOST_FPGA'].bit_index_low)



from ..core import Endpoint
from ..utils import custom_signed_to_int
import numpy as np
import time


class ADCDATA():
    # TODO: add class docstring

    def __init__(self):
        pass

    def deswizzle(self, buf):
        # TODO: add method docstring

        d = np.frombuffer(buf, dtype=np.uint8).astype(np.uint32)
        if self.num_bits == 16:
            d1 = d[0::4] + (d[1::4] << 8)  # TODO: check the byte ordering
            # TODO: is breaking this up necessary? looks like it
            d2 = d[2::4] + (d[3::4] << 8)
        elif self.num_bits == 18:
            # TODO: check 18-bit data conversion
            d2 = (d[3::4] << 8) + d[1::4] + ((d[0::4] << 16) & 0x03)

        if self.name == 'AD7961':
            c = np.empty((d1.size + d2.size,), dtype=d1.dtype)
            # see Xilinx FIFO guide PG057 pg 115, memory fills up from left to right (MSB to LSB)
            c[0::2] = d2
            c[1::2] = d1
        elif self.name == 'ADS8686':
            c = {'A': np.array([]),
                 'B': np.array([])}
            c['A'] = d1
            c['B'] = d2
        return c

    def convert_twos(self, d):
        """Convert data to integer two's complement representation."""

        return custom_signed_to_int(d, self.num_bits)

    def convert_data(self, buf):
        """Deswizle and convert the twos complement representation."""

        return self.convert_twos(self.deswizzle(buf))

    # TODO: test composite ADC function that enables, reads, converts and plots
    # TODO: incorporate/connect QT graphing (UIscript.py)
    def read(self, custom_signed_to_int_conv=True):
        # TODO: add method docstring

        s, e = self.fpga.read_pipe_out(self.endpoints['PIPE_OUT'].address)
        if custom_signed_to_int_conv:
            data = self.convert_data(s)
        else:
            data = self.deswizzle(s)
        return data

    def stream_mult(self, swps=4, custom_signed_to_int_conv=True, data_len=1024):
        # TODO: add method docstring
        print('ADC stream multiple')
        cnt = 0
        st = bytearray(np.asarray(np.ones(0, np.uint8)))
        self.fpga.xem.UpdateTriggerOuts()

        start_time = time.time()
        timeout_time = 1*swps  # seconds
        timeout_flg = (time.time() < (start_time + timeout_time))

        while ((cnt < swps) and timeout_flg):
            # check the FIFO half-full flag
            if (self.fpga.xem.IsTriggered(self.endpoints['FIFO_HALFFULL'].address,
                                          self.endpoints['FIFO_HALFFULL'].bit_index_low)):
                s, e = self.fpga.read_pipe_out(self.endpoints['PIPE_OUT'].address,
                                               data_len)
                st += s
                cnt = cnt + 1
                if self.fpga.debug:
                    print(cnt)
            self.fpga.xem.UpdateTriggerOuts()
            timeout_flg = (time.time() < (start_time + timeout_time))
        if not timeout_flg:
            print('ADC stream_mult timed out')
            return -1
        if custom_signed_to_int_conv:
            data = self.convert_data(st)
        else:
            data = self.deswizzle(st)
        return data



from ..core import Endpoint
from ..utils import test_bit, gen_mask, custom_signed_to_int
import numpy as np
import time
import os
import h5py


class DDR3():
    """
    The DDR is divided into 2 buffers. Each buffer has an incoming and outgoing FIFO.

    **1st buffer:**

    * Function generator like data that provides a data-stream to the DACs
    * Write from the host when DAC_WRITE_ENABLE is set (clear dac read and clear adc read and write)
    * Organized into groups of 16 bits to 8 channels
    * Read to the DACs with set_dac_read
    * 6 AD5453 (14 bits each)
    * 2 DAC80508 (16 bits each) 
    * DAC80508 channel data in the MSBs of the AD5453 stream
    * FIFO: OpalKelly in -> out to DDR [32bits x 1024words = 4096 bytes]
    * FIFO: DDR in -> DAC data out [out: 256bits x 256 = 8192 bytes]

    **2nd buffer:**

    * Buffering for ADC data (AD7961; ADS8686), DAC outgoing data, timestamps.
    * ADC data writes to the DDR set_adc_write()
    * ADC data in DDR can be read to the host with set_adc_read()
    * The DDR througput is sufficient for 1st buffer reading while reading and writing to the 2nd buffer
    * FIFO: OpalKelly in -> out to DDR [in: 32bits x 1024words = 4096 bytes]
    * FIFO: DDR in -> OpalKelly out [256bits x 128 = 4096 bytes] OpalKelly block size is half this.

    **Expected sequence of operations:**

    1) At startup write pattern for DACs using write_channels() [set_dac_write(), clear_dac_write() are called within write_channels()]
    2) set_dac_read() # starts DAC data output to DACs via SPI and then set_adc_write() ADC data captured into DDR
    3) set_adc_read() # allows host to read PipeOut as PipeOut is continuously filled if emptied

    **ADC data ordering and saving:**

    * deswizzle()
    * save_data()

    **DDR configuration bits:**

    * DAC_WRITE_ENABLE
    * DAC_READ_ENABLE
    * ADC_WRITE_ENABLE
    * ADC_TRANSFER_ENABLE
    
    """

    BLOCK_SIZE = 2048  # 1/2 the incoming FIFO depth in bytes (size of the BlockPipeIn)
    # number of channels that the DDR is striped between (for DACs)
    NUM_CHANNELS = 8
    UPDATE_PERIOD = 400e-9  # 2.5 MHz -- requires SCLK ~ 50 MHZ
    PORT1_INDEX = 0x3_7f_ff_f8
    NUM_ADC_CHANNELS = 8  # number of 2 byte chunks in DDR
    ADC_PERIOD = 200e-9

    # the index is the DDR address that the circular buffer stops at.
    # need to write all the way up to this stoping point otherwise the SPI output will glitch
    SAMPLE_SIZE = int((PORT1_INDEX + 8)/4)

    def __init__(self, fpga, endpoints=None, data_version='TIMESTAMPS'):
        if endpoints is None:
            endpoints = Endpoint.get_chip_endpoints('DDR3')
        self.fpga = fpga
        self.endpoints = endpoints
        self.data_version = data_version  # sets deswizzling mode

        self.data_arrays = []
        for i in range(DDR3.NUM_CHANNELS):
            self.data_arrays.append(np.zeros(
                DDR3.SAMPLE_SIZE).astype(np.uint16))

        self.clear_adc_debug()

    def set_adc_debug(self):
        """Set the ADC debug bit.

        That bit multiplexes a counter to ADC channel 0 and bits 47:0 of the
        DAC data to ADC channels 1,2,3.

        Not supported in all versions of the FPGA design.
        """

        self.fpga.set_wire_bit(self.endpoints['ADC_DEBUG'].address,
                               self.endpoints['ADC_DEBUG'].bit_index_low)

    def clear_adc_debug(self):
        """Clear the ADC debug bit.

        DDR ADC data will be from the ADC.
        """
        self.fpga.clear_wire_bit(self.endpoints['ADC_DEBUG'].address,
                                 self.endpoints['ADC_DEBUG'].bit_index_low)

    @staticmethod
    def make_flat_voltage(amplitude):
        """Return a constant unit16 array of value amplitude.

        Array length based on the sample_size parameter. The conversion from
        float or int voltage to int digital (binary) code should take place
        BEFORE this function.

        Parameters
        ----------
        amplitude: int
            Digital (binary) value of the flat voltage

        Returns
        -------
            amplitude : numpy.ndarray 
                to be assigned to DDR data array
        """

        amplitude = np.ones(DDR3.SAMPLE_SIZE)*amplitude
        amplitude = amplitude.astype(np.uint16)
        return amplitude

    @staticmethod
    def closest_frequency(freq):
        """Determine closest frequency so the waveform evenly divides into the length of the DDR3

        Parameters
        ----------
        freq : float
            Desired frequency

        Returns
        -------
        new_frequency : float 
            The closest possible frequency
        """

        samples_per_period = (1/freq) / DDR3.UPDATE_PERIOD

        if samples_per_period <= 2:
            print('Frequency is too high for the DDR update rate')
            return None
        total_periods = DDR3.SAMPLE_SIZE/samples_per_period
        # round and recalculate frequency
        round_total_periods = np.round(total_periods)
        round_samples_per_period = DDR3.SAMPLE_SIZE / \
            round_total_periods
        new_frequency = 1 / \
            (DDR3.UPDATE_PERIOD * round_samples_per_period)

        return new_frequency

    @staticmethod
    def make_sine_wave(amplitude, frequency,
                       offset=0x2000, actual_frequency=True):
        """Return a sine-wave array for writing to DDR.

        The conversion from float or int voltage to int digital (binary) code
        should take place BEFORE this function.

        Parameters
        ----------
        amplitude : int
            Digital (binary) value of the sine wave.
        frequency : float
            Desired frequency in Hz.
        offset : int
            Digital (binary) value offset.
        actual_frequency : bool
            Decide whether closest frequency that fits an integer number of periods is used.

        Returns
        -------
        ddr_seq : numpy.ndarray 
            to be assigned to DDR data array  

        frequency : float
            actual frequency after closest_frequency
        """

        if (amplitude) > offset:
            print('Error: amplitude in sine-wave is too large')
            return -1
        if actual_frequency:
            frequency = DDR3.closest_frequency(frequency)

        t = np.arange(0, DDR3.UPDATE_PERIOD*DDR3.SAMPLE_SIZE,
                      DDR3.UPDATE_PERIOD)
        # print('length of time axis after creation ', len(t))
        ddr_seq = (amplitude)*np.sin(t*frequency*2*np.pi) + offset
        if any(ddr_seq < 0) or any(ddr_seq > (2**16-1)):
            print('Error: Uint16 overflow in make sine wave')
            return -1
        ddr_seq = ddr_seq.astype(np.uint16)
        return ddr_seq, frequency

    @staticmethod
    def make_ramp(start, stop, step, actual_length=True):
        """Create a ramp signal to write to the DDR.

        The conversion from float or int voltage to int digital (binary) code
        should take place BEFORE this function.

        Parameters
        ----------
        start : int
            Digital (binary) value to start the ramp at
        stop : int
            Digital (binary) value to stop the ramp at
        step : int
            Digital (binary) code to step by

        Returns
        -------
            ddr_seq : numpy.ndarray 
                to be assigned to DDR data array        
        """

        # change the stop value for integer number of cycles
        ramp_seq = np.arange(start, stop, step)
        len_ramp_seq = len(ramp_seq)
        if actual_length:
            length = int(
                DDR3.SAMPLE_SIZE/np.round(DDR3.SAMPLE_SIZE/len_ramp_seq))
            stop = start + length*step
            ramp_seq = np.arange(start, stop, step)
        num_tiles = DDR3.SAMPLE_SIZE//len(ramp_seq)
        extras = DDR3.SAMPLE_SIZE % len(ramp_seq)
        ddr_seq = np.tile(ramp_seq, num_tiles)
        ddr_seq = np.hstack((ddr_seq, ramp_seq[0:extras]))
        ddr_seq = ddr_seq.astype(np.uint16)
        return ddr_seq

    @staticmethod
    def make_step(low, high, length, actual_length=True, duty=50):
        """Return a step signal (square wave) to write to the DDR.

        The conversion from float or int voltage to int digital (binary) code
        should take place BEFORE this function.

        Parameters
        ----------
        low : int
            Digital (binary) code for the low value of the step.
        high : int
            Digital (binary) code for the high value of the step.
        length : TODO add type for length
            TODO add description for length
        actual_length : bool
            TODO add description for actual_length
        duty : int or float
            Duty cycle percentage. Enter as a percentage [0.0, 100.0].

        Returns
        -------
        ddr_seq : numpy.ndarray 
            to be assigned to DDR data array  
        """

        if actual_length:
            length = int(
                DDR3.SAMPLE_SIZE/np.round(DDR3.SAMPLE_SIZE/length))
        l_first = int(length/100*duty)
        l_end = int(length/100*(100-duty))
        ramp_seq = np.concatenate((np.ones(l_first)*low, np.ones(l_end)*high))
        num_tiles = DDR3.SAMPLE_SIZE//len(ramp_seq)
        extras = DDR3.SAMPLE_SIZE % len(ramp_seq)
        ddr_seq = np.tile(ramp_seq, num_tiles)
        ddr_seq = np.hstack((ddr_seq, ramp_seq[0:extras]))
        ddr_seq = ddr_seq.astype(np.uint16)
        return ddr_seq

    def write_channels(self, set_ddr_read=True):
        """Write the channels as striped data to the DDR."""

        data = np.zeros(
            int(len(self.data_arrays[0])*DDR3.NUM_CHANNELS))
        data = data.astype(np.uint16)

        for i in range(DDR3.NUM_CHANNELS):
            if i % 2 == 0:  # extra order swap on the 32 bit wide pipe
                data[(7-i - 1)::8] = self.data_arrays[i]
            else:
                data[(7-i + 1)::8] = self.data_arrays[i]

        print('Length of data DDR data [2 byte words] = {}'.format(len(data)))
        return self.write_buf(bytearray(data), set_ddr_read=set_ddr_read)

    def write_buf(self, buf, set_ddr_read=True):
        """Write a bytearray to the DDR3.

        Parameters
        ----------
        buf : bytearray
            bytearray to write to the DDR

        Returns
        -------
        block_pipe_return : int
            length of the buffer written to the DDR (or error code if unsuccessful)
        speed_MBs : float
            speed of the write in MB/s
        """

        print('Length of buffer being written to DDR [bytes]: ', len(buf))
        self.clear_dac_read()
        self.reset_fifo(name='DAC_IN')
        self.set_dac_write()

        print('Writing to DDR...')
        time1 = time.time()
        block_pipe_return = self.fpga.xem.WriteToBlockPipeIn(epAddr=self.endpoints['BLOCK_PIPE_IN'].address,
                                                             blockSize=DDR3.BLOCK_SIZE,
                                                             data=buf)
        print(f'The length of the DDR write was {block_pipe_return}')

        time2 = time.time()
        time3 = (time2-time1)
        speed_MBs = (int)(block_pipe_return/1024/1024/time3)
        print(f'The speed of the write was {speed_MBs} MB/s')

        # below prepares the HDL into read mode
        self.clear_dac_write()
        if set_ddr_read:
            self.set_dac_read()
            self.set_adc_read()
        return block_pipe_return, speed_MBs

    def reset_mig_interface(self):
        """Reset user interface to the MIG (memory interface generator). Resets
            the DDR address pointers for read/write of both buffers 
            (does not reset the MIG controller).
        """
        self.fpga.send_trig(self.endpoints['UI_RESET'])

    def fifo_status(self):
        """Check the empty, full, and count status of the DDR interfacing FIFOs

        Returns
        -------
        fifo_status : dict 
            dictionary of fifo status ('EMPTY', 'FULL', 'ADC_DATA_COUNT') for 'IN', 'OUT', and channels 1, 2
        """
        wire_status = self.fpga.read_wire(
            self.endpoints['INIT_CALIB_COMPLETE'].address)
        fifo_status = {}
        for fe in ['EMPTY', 'FULL']:
            for num in [1, 2]:
                for inout in ['IN', 'OUT']:
                    ep_name = '{}{}_{}'.format(inout, num, fe)
                    bit_pos = self.endpoints[ep_name].bit_index_low
                    val = test_bit(wire_status, bit_pos)
                    fifo_status[ep_name] = val
                    print('{} = {}'.format(ep_name, val))

        ep_name = 'ADC_DATA_COUNT'
        cnt_bit_low = self.endpoints[ep_name].bit_index_low
        cnt_bit_width = self.endpoints[ep_name].bit_width
        cnt_msk = gen_mask(np.arange(cnt_bit_low, cnt_bit_width))
        fifo_status[ep_name] = (wire_status & cnt_msk) >> cnt_bit_low
        print('{} = {}'.format(ep_name, val))

        # self.print_fifo_status(fifo_status)
        return fifo_status

    def print_fifo_status(self, fifo_status):
        """Print the FIFO status dictionary.

        Parameters
        ----------
        fifo_status : dict
            Dictionary of fifo status.
        """

        for k in fifo_status:
            print('{} = {}'.format(k, fifo_status[k]))

    def set_dac_read(self):
        """Set DDR / FIFOs read enable. Enables DDR data going to the 
        DACs and ADC data into DDR
        """

        self.fpga.set_wire_bit(self.endpoints['DAC_READ_ENABLE'].address,
                               self.endpoints['DAC_READ_ENABLE'].bit_index_low)

    def clear_dac_read(self):
        """Clear DDR / FIFOs read enable. Stops DDR data from going to 
        the DACs and ADC data into DDR
        """

        self.fpga.clear_wire_bit(self.endpoints['DAC_READ_ENABLE'].address,
                                 self.endpoints['DAC_READ_ENABLE'].bit_index_low)

    def set_dac_write(self):
        """Set DDR / FIFOs write enable into DDR via Pipe.
        """

        self.fpga.set_wire_bit(self.endpoints['DAC_WRITE_ENABLE'].address,
                               self.endpoints['DAC_WRITE_ENABLE'].bit_index_low)

    def clear_dac_write(self):
        """Clear DDR / FIFOs write enable into DDR via Pipe.
        """

        self.fpga.clear_wire_bit(self.endpoints['DAC_WRITE_ENABLE'].address,
                                 self.endpoints['DAC_WRITE_ENABLE'].bit_index_low)

    def set_adc_write(self):
        """Set DDR / FIFOs write enable ADC data into DDR.
        """

        self.fpga.set_wire_bit(self.endpoints['ADC_WRITE_ENABLE'].address,
                               self.endpoints['ADC_WRITE_ENABLE'].bit_index_low)

    def set_adc_dac_simultaneous(self):
        """
        Set DDR / FIFOs read enable. Enables DDR data going to the 
        DACs and ADC data into DDR
        """
        bit_addr = [self.endpoints['ADC_WRITE_ENABLE'].bit_index_low,
                    self.endpoints['DAC_READ_ENABLE'].bit_index_low,
                    self.endpoints['ADC_TRANSFER_ENABLE'].bit_index_low]
        bit_vals = [1, 1, 1]

        self.fpga.set_ep_simultaneous(self.endpoints['ADC_WRITE_ENABLE'].address,
                                      bit_addr, bit_vals)

    def clear_adc_write(self):
        """Clear DDR / FIFOs write enable ADC data into DDR.
        """

        self.fpga.clear_wire_bit(self.endpoints['ADC_WRITE_ENABLE'].address,
                                 self.endpoints['ADC_WRITE_ENABLE'].bit_index_low)

    def set_adc_read(self):
        """Set DDR / FIFO read enable DDR data from the ADCs out via a PipeOut.
        """
        self.fpga.set_wire_bit(self.endpoints['ADC_TRANSFER_ENABLE'].address,
                               self.endpoints['ADC_TRANSFER_ENABLE'].bit_index_low)

    def clear_adc_read(self):
        """Clear DDR / FIFO read enable DDR data from the ADCs out via a PipeOut.
        """

        self.fpga.clear_wire_bit(self.endpoints['ADC_TRANSFER_ENABLE'].address,
                                 self.endpoints['ADC_TRANSFER_ENABLE'].bit_index_low)

    def set_adcs_connected(self):
        """Set that AD7961s are connected to the FPGA so that DDR3 write enable comes from 
            the AD7961.v module.
        """
        self.fpga.set_wire_bit(self.endpoints['USE_ADC_READY'].address,
                               self.endpoints['USE_ADC_READY'].bit_index_low)

    def clear_adcs_connected(self):
        """ AD7961s are not connected to the FPGA so the DDR3 write enable signal is emulated
            by the global_timing module.
        """
        self.fpga.clear_wire_bit(self.endpoints['USE_ADC_READY'].address,
                                 self.endpoints['USE_ADC_READY'].bit_index_low)

    def reset_fifo(self, name):
        """Reset FIFO interfaces to DDR .
        """
        fifo_reset_names = ['DAC_IN', 'DAC_READ', 'ADC_IN', 'ADC_TRANSFER']
        if (name not in fifo_reset_names) and (not name == 'ALL'):
            print(
                f'DDR3 FIFO reset name of {name} is not in list of {fifo_reset_names}')

        else:
            if name == 'ALL':
                for name_tmp in fifo_reset_names:
                    ep_name = 'FIFO_' + name_tmp + '_RST'
                    self.fpga.set_wire_bit(self.endpoints[ep_name].address,
                                           self.endpoints[ep_name].bit_index_low)
                    self.fpga.clear_wire_bit(self.endpoints[ep_name].address,
                                             self.endpoints[ep_name].bit_index_low)
            else:
                ep_name = 'FIFO_' + name + '_RST'
                self.fpga.set_wire_bit(self.endpoints[ep_name].address,
                                       self.endpoints[ep_name].bit_index_low)
                self.fpga.clear_wire_bit(self.endpoints[ep_name].address,
                                         self.endpoints[ep_name].bit_index_low)

    def adc_single(self):
        """Set ADC read address to the ADC write address.

        Emulates an immediate "trigger" of an oscilloscope.
        """

        self.fpga.set_wire_bit(self.endpoints['ADC_ADDR_SET'].address,
                               self.endpoints['ADC_ADDR_SET'].bit_index_low)
        self.fpga.send_trig(self.endpoints['ADC_ADDR_RESET'])

    def read_adc_block(self, sample_size=None, source='ADC', DEBUG_PRINT=False):
        """Read ADC (and other) DDR data. 
        Block size must be a power of two from 16 to 16384
        will automatically perform multiple transfers to complete the full LENGTH.
        The length must be an integer multiple of 16 for USB3.0
        and the length must be an Integer multiple of Block Size.
        see https://docs.opalkelly.com/fpsdk/frontpanel-api/ section 3.3.1

        Parameters
        ----------
        sample_size : int
            Length of read in bytes. If none uses DDR parameter 'sample_size.'

        source : str
            FIFO output buffer to read. Either 'ADC' or 'FG'. 'FG' just reads
            back what is written for DACs (as function generator) so not so
            useful.

        Returns
        -------
        data_buf : byearray
            adc data read as a bytearray
        read_cnt : int
            The count (or error code) read from the OpalKelly interface
        """

        if sample_size is None:
            data = np.zeros((DDR3.SAMPLE_SIZE,), dtype=int)
            data_buf = bytearray(data)
        else:
            data_buf = bytearray(sample_size)

        block_size = DDR3.BLOCK_SIZE
        # check block size
        if block_size % 16 != 0:
            print('Error in read adc. Block size is not a multiple of 16')
            return -1, -1
        if block_size > 16384:
            print('Error in read adc. Block size is greater than 16384')
            return -2, -2

        if source == 'ADC':
            read_cnt = self.fpga.xem.ReadFromBlockPipeOut(epAddr=self.endpoints['BLOCK_PIPE_OUT'].address,
                                                          blockSize=block_size,
                                                          data=data_buf)

        elif source == 'FG':  # TODO: test, don't expect this to work
            read_cnt = self.fpga.xem.ReadFromBlockPipeOut(epAddr=self.endpoints['BLOCK_PIPE_OUT_FG'].address,
                                                          blockSize=block_size,
                                                          data=data_buf)
        else:
            print('Incorrect source in read_adc')
            return -11, -11

        if DEBUG_PRINT:
            print(
                f'The length [num of bytes] of the BlockPipeOut read is: {read_cnt}')
        return data_buf, read_cnt

    def deswizzle(self, d, convert_twos=True):
        """Reorder DDR data to match the ADC channels. Shift MSBytes up by 8 
        and combine with LSBytes. Swap channels to match ADC channel numbering.

        Parameters
        ----------
        d : array
            array of bytes. 
        convert_twos : Boolean
            if true converts data to signed

        Returns
        -------
        chan_data : dict
            dictionary of data arrays (keys are channel numbers)

        """
        bits = 16
        chan_data_swz = {}  # this data is swizzled

        # first version of ADC data before DACs + timestamps are stored
        if self.parameters['data_version'] == 'ADC_NO_TIMESTAMPS':
            for i in range(4):
                chan_data_swz[i] = (d[(0 + i * 2):: 8] << 0) + \
                    (d[(1 + i * 2):: 8] << 8)

            chan_data = {}
            chan_data[0] = chan_data_swz[2]
            chan_data[1] = chan_data_swz[3]
            chan_data[2] = chan_data_swz[0]
            chan_data[3] = chan_data_swz[1]
            if convert_twos:
                for i in range(4):
                    chan_data[i] = custom_signed_to_int(chan_data[i], bits)

        # first version of ADC data before DACs + timestamps are stored
        if self.parameters['data_version'] == 'TIMESTAMPS':
            for i in range(8):
                chan_data_swz[i] = (d[(0 + i * 2):: 16] << 0) + \
                    (d[(1 + i * 2):: 16] << 8)

            chan_data = {}
            chan_data[0] = chan_data_swz[6]
            chan_data[1] = chan_data_swz[7]
            chan_data[2] = chan_data_swz[5]
            chan_data[3] = chan_data_swz[4]

            chan_data[4] = chan_data_swz[2]
            chan_data[5] = chan_data_swz[3]
            chan_data[6] = chan_data_swz[0]
            chan_data[7] = chan_data_swz[1]

        return chan_data

    def data_to_names(self, chan_data, bitfile_version=None):
        """
        Put deswizzled data into dictionaries with names that match with the data sources. 
        Complete twos complement conversion where necessary. Check timestamps for skips.
        Check the constant values for errors.

        This supports 2 versions of the FPGA code:
        'ADC_NO_TIMESTAMPS': DDR data is only the fast ADC. AD7961
        'TIMESTAMPS': DDR data is numerous. AD7961, AD5453 out, ADS8686, timestamps, readcheck


        Parameters
        ----------
        chan_data : dict of np.arrays
            data from reading DDR (minimally processed into 2 byte containers)
        bitfile_version : int
            The bitfile version that the .h5 file which held chan_data was created with. Defaults to the bitfile version of the FPGA
        
        Returns
        -------
        adc_data : dict 
            fast adc data (double format @ 5 MSPS)
        timestamp : np.array  
            timestamps
        dac_data : dict 
            DAC output data
        ads : dict 
            ADS8686 ADC data (double format @ 1 MSPS)

        ads_seq_cnt : dict
            two keys 0, 1. Each value is an array of sequence counts (counts from 0 to 23)
            the first key will have 0,2,4,6, ... , 22 (or similar)
            the second key will have 1,3,5, ..., 23 (or similar)

        error : boolean 
            if True the constant read values were wrong 
               or the timestamp steps are not all the same
        """

        if bitfile_version is None:
            # Old code, hasn't been updated to pass bitfile_version from .h5 file header
            # Default to FPGA bitfile version, or version 1 if the FPGA bitfile version is None
            if self.fpga.bitfile_version is None:
                bitfile_version = 1
            else:
                bitfile_version = self.fpga.bitfile_version

        # first version of ADC data before DACs + timestamps are stored
        if self.parameters['data_version'] == 'ADC_NO_TIMESTAMPS':
            adc_data = chan_data
            timestamp = np.nan
            read_check = np.nan
            dac_data = np.nan
            ads = np.nan

        # first version of ADC data before DACs + timestamps are stored
        if self.parameters['data_version'] == 'TIMESTAMPS':

            adc_data = {}
            for i in range(4):
                # adc_data[i] = custom_signed_to_int(chan_data[i], 16)
                adc_data[i] = chan_data[i]

            if bitfile_version < 2: # 00.00.02 -> 2
                lsb = chan_data[6][0::5].astype(np.uint64)
            else:
                lsb = chan_data[7][1::5].astype(np.uint64)
            mid_b = ((chan_data[6][1::5].astype(np.uint64)) << 16)
            msb = ((chan_data[7][2::5].astype(np.uint64)) << 32)
            t_len = np.size(msb)
            timestamp = (lsb[0:(t_len-1)] +
                         mid_b[0:(t_len-1)] + msb[0:(t_len-1)])

            read_check = {}
            read_check[0] = chan_data[7][3::10]
            read_check[1] = chan_data[7][4::10]
            read_check[2] = chan_data[7][8::10]
            read_check[3] = chan_data[7][9::10]

            ads_seq_cnt = {}
            ads_seq_cnt[0] = (chan_data[7][4::10] & 0x001f)
            ads_seq_cnt[1] = (chan_data[7][9::10] & 0x001f)

            dac_data = {}
            dac_data[0] = chan_data[4][0::2]
            dac_data[1] = chan_data[4][1::2]
            dac_data[2] = chan_data[5][0::2]
            dac_data[3] = chan_data[5][1::2]
            # dac channels 4,5 are available but not every sample. skip for now. TODO: add channels 4,5

            ads = {}
            ads['A'] = custom_signed_to_int(chan_data[7][0::5], 16)
            if bitfile_version < 2: # 00.00.02 -> 2
                ads['B'] = custom_signed_to_int(chan_data[7][1::5], 16)
            else:
                ads['B'] = custom_signed_to_int(chan_data[6][0::5], 16)
            error = False
            # check that the constant values are constant
            constant_values = {0: 0xaa55, 1: (0x28b<<5), 2: 0x77bb, 3: (0x28c<<5)}
            constant_value_mask = {0: 0xffff, 1: 0xffe0, 2: 0xffff, 3: 0xffe0}
            for i in range(4):
                if not np.all( (read_check[i] & constant_value_mask[i]) == constant_values[i]):
                    print(f'Error in constant value: {constant_values[i]} ')
                    print(
                        f'Number of errors: {np.sum(read_check[i] != constant_values[i])}')
                    error = True

            # check the timestamps for a skip
            unq_time_intervals = np.unique(np.diff(timestamp))
            if np.size(unq_time_intervals) > 1:
                print('Warning: Multiple time intervals')
                error = True

            return adc_data, timestamp, dac_data, ads, ads_seq_cnt, error

    def save_data(self, data_dir, file_name, num_repeats=4, blk_multiples=40, append=False):
        """
        read and save DDR data to an hdf file 

        Parameters
        ----------
        data_dir : string
            directory for data
        file_name : string
            filename to save data (does not append extension).
        num_repeats : int
            total data read is 2048 bytes * num_repeats * blk_multiples
        blk_multiples : int 
            number of blocks read by adc_read. adc_read is a single OpalKelly API call 

        Returns
        -------
        new_data : np.ndarray
            The new data saved in the h5 file.
        """

        # If the file doesn't already exist, write a new one
        full_data_name = os.path.join(data_dir, file_name)
        if append:
            if not os.path.exists(full_data_name):
                print(f'No existing file found at {full_data_name}, creating new file')
                append = False
        else:
            try:
                os.remove(full_data_name)
            except OSError:
                pass

        if append:
            file_mode = 'a'
        else:
            file_mode = 'w'

        chunk_size = int(DDR3.BLOCK_SIZE * blk_multiples / (
            DDR3.NUM_ADC_CHANNELS*2))  # readings per ADC
        repeat = 0
        adc_readings = chunk_size*num_repeats
        print(f'Anticipated chunk size (readings per channel) {chunk_size}')
        print(
            f'Reading {adc_readings*2/1024} kB per ADC channel for a total of {adc_readings*DDR3.ADC_PERIOD*1000} ms of data')

        self.set_adc_read()  # enable data into the ADC reading FIFO
        time.sleep(adc_readings*DDR3.ADC_PERIOD)

        # Save ADC DDR data to a file
        with h5py.File(full_data_name, file_mode) as file:
            if append:
                data_set = file['adc']
                new_data_index = data_set.shape[1]
                if data_set.attrs['bitfile_version'] != self.fpga.bitfile_version:
                    raise Exception(f"File {os.path.join(data_dir, file_name)} bitfile version {data_set.attrs['bitfile_version']} does not match FPGA bitfile version {self.fpga.bitfile_version}")
                # Make space for first round of new data. Not needed when not appending because the data set is created with chunk_size space
                data_set.resize(data_set.shape[1] + chunk_size, axis=1)
            else:
                data_set = file.create_dataset("adc", (DDR3.NUM_ADC_CHANNELS, chunk_size), maxshape=(
                    DDR3.NUM_ADC_CHANNELS, None))
                data_set.attrs['bitfile_version'] = self.fpga.bitfile_version
                new_data_index = 0

            while repeat < num_repeats:
                d, bytes_read_error = self.read_adc(blk_multiples)
                if self.parameters['data_version'] == 'ADC_NO_TIMESTAMPS':
                    chan_data = self.deswizzle(d)

                elif self.parameters['data_version'] == 'TIMESTAMPS':
                    chan_data = self.deswizzle(d)

                if DDR3.NUM_ADC_CHANNELS == 4:
                    chan_stack = np.vstack(
                        (chan_data[0], chan_data[1], chan_data[2], chan_data[3]))

                if DDR3.NUM_ADC_CHANNELS == 8:
                    chan_stack = np.vstack((chan_data[0], chan_data[1], chan_data[2], chan_data[3],
                                            chan_data[4], chan_data[5], chan_data[6], chan_data[7]))

                repeat += 1
                if repeat == 0:
                    print(f'Chunk size by chan data size {chunk_size}')
                    data_set[:] = chan_stack
                else:
                    data_set[:, -chunk_size:] = chan_stack
                if repeat < num_repeats:
                    data_set.resize(data_set.shape[1] + chunk_size, axis=1)
            new_data = data_set[:, new_data_index:]

        print(f'Done with DDR reading: saved as {full_data_name}')
        return new_data

    def read_adc(self, blk_multiples=2048):
        """ 
        read DDR data into a numpy buffer of bytes

        Parameters
        ----------
        blk_multiples : int
            total size of the read is blk_multiples * block_size  

        Returns
        -------
        d : bytearray
            data as uint32    
        bytes_read_error : int 
            bytes read or error code          
        """

        t, bytes_read_error = self.read_adc_block(  # just reads from the block pipe out
            sample_size=DDR3.BLOCK_SIZE * blk_multiples
        )
        d = np.frombuffer(t, dtype=np.uint8).astype(np.uint32)
        print(f'Bytes read: {bytes_read_error}')
        return d, bytes_read_error

    def set_index(self, factor, factor2=None):
        """
        No longer used. Index (the DDR address that wraps-around to 0).
        This is now fixed to improve timing performance.
        """
        print('Set index is no longer used. Index is fixed to: {}'.format(
            DDR3.PORT1_INDEX))

    def write_setup(self, data_driven_clock=True):
        """Set up DDR for writing."""

        if data_driven_clock:
            self.set_adcs_connected()
        else:
            self.clear_adcs_connected()
        self.clear_dac_read()
        self.clear_adc_write()
        self.clear_adc_read()    # Stop putting data in outgoing FIFO for Pipe read
        self.reset_fifo(name='ALL')
        self.reset_mig_interface()

    def repeat_setup(self):
        """Setup for reading new data without writing to the DDR again."""

        # stop access to the FIFOs so that after reset of the FIFO(s) no new data is added/extracted
        self.clear_adc_read()
        self.clear_adc_write()
        self.clear_dac_read()
        self.reset_fifo(name='ALL')
        # self.fpga.send_trig(self.endpoints['UI_RESET'])
        self.reset_mig_interface()
        # note that the MIG interface addresses are driven by the FIFOs so will idle
        # until the FIFOs are reenable with write_finish()
        self.write_finish()
        time.sleep(0.01)

    def write_finish(self):
        # reenable both DACs
        self.set_adc_dac_simultaneous()  # enable DAC playback and ADC writing to DDR



from ..core import Register
from ..utils import int_to_list
from .I2CController import I2CController

class UID_24AA025UID(I2CController):
    """Class for the ID chip 24AA025UID.

    Subclass of the I2CController class. Attributes and methods below are
    differences in this class from I2CController only.

    Attributes
    ----------
    ADDRESS_HEADER : int
        7-bit device address with R/W bit space and 4 MSBs filled in specific
        to this chip. The rest is left as 0 to be filled in later.
    registers : dict
        Name-Register pairs for the internal registers of the TCA9555 chip.
    addr_pins : int
        3 LSBs of the 7-bit device address formed alongside the address header
        used to differentiate between different instances of the TCA9555 chip.
    """

    ADDRESS_HEADER = 0b10100000
    registers = Register.get_chip_registers('24AA025UID')

    def write(self, data, word_address=0x00, num_bytes=None):
        """Write data into memory at word_address."""

        dev_addr = UID_24AA025UID.ADDRESS_HEADER | (self.addr_pins << 1)
        print(
            f'word_address: {hex(word_address)}\ndata: {hex(data)}\nnum_bytes: {num_bytes}')

        # Convert data into a list
        if num_bytes is None:
            # User did not specify a number of bytes
            list_data = int_to_list(integer=data)
            num_bytes = len(list_data)
        else:
            # User specified a number of bytes
            list_data = int_to_list(integer=data, num_bytes=num_bytes)

        if list_data is False:
            # int_to_list returns False when the data is longer than the number of bytes
            print('ERROR: data exceeds given number of bytes')
            return False

        if num_bytes > 16:
            # After 16 bytes (1 page), the UID chip will rollover and overwrite data from ealier.
            # To fix this, we start a new I2C write command at the next page
            self.i2c_write_long(devAddr=dev_addr, regAddr=[
                                word_address], data_length=16, data=list_data[:16])
            # We can recursively call this function instead of looping through multiple times.
            # recurse_data = data % (0x1 << (8 * (num_bytes - 16)))
            recurse_data = data
            for i in range(16):
                recurse_data -= list_data[i] << (8 * i)
            recurse_data >>= 16 * 8
            # print(f'{data} % {(0x1 << (8 * (num_bytes - 16)))} = {recurse_data}')
            self.write(data=recurse_data, word_address=(word_address
                       + 16) % 0x80, num_bytes=num_bytes - 16)
            return True

        # Data is no more than 16 bytes (1 page) and can be written with 1 I2C command.
        self.i2c_write_long(devAddr=dev_addr, regAddr=[
                            word_address], data_length=num_bytes, data=list_data)
        return True

    def read(self, word_address=0x00, words_read=1):
        """Return words_read words of data from memory at word_address."""

        # A word is a byte, 8 bits
        dev_addr = UID_24AA025UID.ADDRESS_HEADER | (
            self.addr_pins << 1) | 0b1
        # Bug when reading more than 64 bytes, only 128 bytes possible so check and split to 2 operations
        if words_read > 64:
            d0 = self.i2c_read_long(dev_addr, [word_address], 64)
            d1 = self.i2c_read_long(
                dev_addr, [word_address + 64], words_read - 64)
            return d0 + d1
        return self.i2c_read_long(dev_addr, [word_address], words_read)

    def get_serial_number(self):
        """Return the unique 32-serial number stored in memory on the chip."""

        serial_list = self.read(
            word_address=UID_24AA025UID.registers['SERIAL_NUMBER'].address, words_read=4)
        serial_number = 0
        for byte in serial_list:
            serial_number <<= 8
            serial_number |= byte

        return serial_number

    def get_manufacturer_code(self):
        """Return the chip's manufacturer code.

        For the chips we are using, it should be 0x29.
        """

        return self.read(word_address=UID_24AA025UID.registers['MANUFACTURER_CODE'].address)[0]

    def get_device_code(self):
        """Return the chip's device code.

        For the chips we are using, it should be 0x41.
        """

        return self.read(word_address=UID_24AA025UID.registers['DEVICE_CODE'].address)[0]



from ..core import Endpoint
from ..utils import gen_mask
from .ADCDATA import ADCDATA
import copy
import time

class AD7961(ADCDATA):
    """An interface to the AD7961 ADC
    Passed the FPGA object
    Allows for multiple channels
    read status flags (FIFO full, empty, count and PLL)
    configures FPGA internal resets
    configures chip specific and global enables
    Formats data returned from the wire out
    """
    # TODO: add Attributes to docstring

    # the AD7961 does not have internal registers -- just OK endpoints

    def __init__(self, fpga, endpoints=None):
        if endpoints is None:
            endpoints = Endpoint.get_chip_endpoints('AD7961')
        self.fpga = fpga
        self.endpoints = endpoints
        self.name = 'AD7961'
        self.num_bits = 16  # for AD7961 bits=18 for AD7960

    @staticmethod
    def create_chips(fpga, number_of_chips, endpoints=None):
        """Instantiate a given number of new chips.

        We increment the endpoints between each instantiation as well. The
        number must be greater than 0. If the endpoints argument is left
        as None, then we will use copies of the endpoints_from_defines
        dictionary for the endpoints for each instance, and update that
        original dictionary when we increment the endpoints. This way, the
        endpoints there are ready for another instantiation if needed.
        """

        if type(number_of_chips) is not int or number_of_chips <= 0:
            print('number_of_chips must be an integer greater than 0')
            return False

        if endpoints is None:
            endpoints = Endpoint.endpoints_from_defines.get('AD7961')

        chips = []
        for i in range(number_of_chips):
            # Use deepcopy here to keep the endpoints for different instances separate
            chips.append(AD7961(fpga=fpga, endpoints=copy.deepcopy(endpoints)))
            Endpoint.advance_endpoints(endpoints)
        return chips

    def get_status(self):
        """ Get AD796x status:
        pll lock, FIFO full, half-full, and empty

        Returns: dictionary of status
        """
        status = self.get_fifo_status()
        pll_lock = self.get_pll_status()
        status['pll_lock'] = pll_lock
        for k in status:
            print('{} status of {} = {}'.format(self.name,
                                                k, status[k]))
        return status

    def get_pll_status(self):
        """Get status of the phase locked loop."""

        return self.fpga.read_wire_bit(self.endpoints['PLL_LOCKED'].address,
                                       self.endpoints['PLL_LOCKED'].bit_index_low)

    def get_timing_pll_status(self):
        """Get status of the timing of the phase locked loop."""

        return self.fpga.read_wire_bit(self.endpoints['TIMING_PLL_LOCKED'].address,
                                       self.endpoints['TIMING_PLL_LOCKED'].bit_index_low)

    def get_fifo_status(self):
        """Get fullness status of the FIFO."""

        flags = ['FULL', 'HALFFULL', 'EMPTY']
        fifo_status = {}
        self.fpga.xem.UpdateTriggerOuts()
        for k in flags:
            fifo_status[k] = self.fpga.xem.IsTriggered(self.endpoints['FIFO_{}'.format(k)].address,
                                                       self.endpoints['FIFO_{}'.format(k)].bit_index_low)
        return fifo_status

    # Enable pins:
    # 0000 - power down
    # 1001 - enabled with ref buffer on; 28 MHZ
    # 0100 - test patterns on LVDS
    # 1101 - enabled with ref buffer on; 9 MHZ
    # Note that EN[3]=1 enables the VCM output buffer.

    def power_down_adc(self):
        """Power down single channel of the ADC.

        Fails if the global enables are 010 for LVDS test patterns
        """

        self.set_enables(0b0, global_enables=False)

    def test_pattern(self):
        """Enable PRBS test pattern on the LVDS interface."""

        self.set_enables(value=0b0100)

    def power_up_adc(self, bw='28M'):
        # TODO: add method docstring
        # TODO: explain what bw should be in docstring

        if bw == '28M':
            self.set_enables(0b1001)
        elif bw == '9M':
            self.set_enables(0b1101)
        else:
            print(f'incorrect input sampling bandwidth for {self.name}')

    def power_down_all(self):
        """Power down all channels of the ADC."""

        self.set_enables(0b0000)  # TODO: return anything here?

    def reset_pll(self):
        """Reset the phase locked loop."""

        return self.fpga.xem.ActivateTriggerIn(self.endpoints['PLL_RESET'].address,
                                               self.endpoints['PLL_RESET'].bit_index_low)

    def reset_trig(self):
        """Reset the FPGA controller for the ADC.
        
        Common to synchronize timing (resets ad7961_timing module AND the
        ads8686 timing) 
        """

        return self.fpga.xem.ActivateTriggerIn(self.endpoints['RESET'].address,
                                               self.endpoints['RESET'].bit_index_low)

    def reset_wire(self, value):
        """Set the value of the wire to reset the FPGA controller for the ADC.

        Uses the Opal Kelly WireIn (the trigger can't hold the reset).

        Parameters
        -----------
        value : int
            value = 1 is reset
            value = 0 releases reset
        """

        if value == 1:
            self.fpga.set_wire_bit(self.endpoints['WIRE_RESET'].address,
                                   self.endpoints['WIRE_RESET'].bit_index_low)
        if value == 0:
            self.fpga.clear_wire_bit(self.endpoints['WIRE_RESET'].address,
                                     self.endpoints['WIRE_RESET'].bit_index_low)

    def power_down_fpga(self):
        """Power down the FPGA controller through WireIn."""

        # TODO: add functionality (wirein bit) to the FPGA
        #       this would reduce FPGA power consumption
        self.reset_wire(0)

    def set_enables(self, value=0b0000, global_enables=True):
        """Set the EN0 specific to this channel (LSB in values).

        Optionally also modify the global enables (all channels).
        """

        mask = gen_mask(self.endpoints['ENABLE'].bit_index_low)
        if global_enables:
            # global enables (connect to all AD7961); create a list of bit position
            gl_mask = [x+self.endpoints['GLOBAL_ENABLE'].bit_index_low
                       for x in range(self.endpoints['GLOBAL_ENABLE_LEN'].bit_index_low)]
            # or global mask with the signal channel EN0 mask
            mask = gen_mask(gl_mask) | mask

        # setup the values
        value_chan = (value & 0b0001) << (
            self.endpoints['ENABLE'].bit_index_low)
        if global_enables:
            #  globabl enables are the 3 MSBs
            val_global = (value & 0b1110) << (
                self.endpoints['GLOBAL_ENABLE'].bit_index_low - 1)
            # minus 1 since already left shifted by 1
            print('Global value = 0x{:0x}'.format(val_global))
            value = value_chan | val_global

        print('Enables setting value = 0x{:0x} with mask = = 0x{:0x}'.format(
            value, mask))
        self.fpga.set_wire(self.endpoints['ENABLE'].address, value, mask=mask)

    def setup(self, reset_pll=False):
        """
        0) put ADC controller into reset
        1) optionally reset the ADC PLL and wait for lock
            (only need to reset PLL on first channel)
        2) power up adc
        3) release ADC controller reset
        4) reset fifo
        5) check and return status (PLL and fifo)
        """
        self.reset_wire(1)
        if reset_pll:
            # reset PLL
            self.reset_pll()
            pll_cnt = 0
            while not self.get_pll_status():
                time.sleep(0.01)
                pll_cnt += 1
                if pll_cnt > 20:
                    print('PLL lock timeout')
                    return -1
        # enable ADC
        self.power_up_adc()
        # now that PLL is locked and ADC is ready
        # release the reset of the ADC controller
        # self.reset_trig()
        self.reset_wire(0)  # releases the reset
        # reset FIFO
        self.reset_fifo()
        status = self.get_status()
        # TODO: the FIFO is guaranteed to overflow
        return status



import numpy as np
import time


class I2CController:
    """Class for controllers on the FPGA using I2C protocol.

    Attributes
    ----------
    I2C_MAX_TIMEOUT_MS : int
        Maximum wait time until transmission timeout in milliseconds.
    i2c : dict
        Dictionary of I2C memory buffer and data start location.
    fpga : FPGA
        FPGA instance this controller uses to communicate.
    endpoints : dict
        Endpoints on the FPGA this controller uses to communicate.
    """

    I2C_MAX_TIMEOUT_MS = 50

    def __init__(self, fpga, addr_pins, endpoints, i2c={'m_pBuf': [], 'm_nDataStart': 7}):
        self.i2c = i2c
        self.fpga = fpga
        self.addr_pins = addr_pins
        self.endpoints = endpoints

    @classmethod
    def create_chips(cls, fpga, addr_pins, endpoints):
        """Instantiate a number of new I2C chips.

        The FPGA and endpoints will be the same for all instantiated chips.

        Parameters
        ----------
        fpga : FPGA
            The fpga instance for the chip to connect with.
        addr_pins : list
            The list of addr_pins assigned to the new chips.

        Returns
        -------
        list
            A list of the newly instantiated chips in the same order addr_pins was given in.
        """

        return [cls(fpga=fpga, addr_pins=addr, endpoints=endpoints) for addr in addr_pins]

    # STARTS - Defines the preamble bytes after which a start bit is
    #      transmitted. For example, if STARTS=0x04, a start bit is
    #      transmitted after the 3rd preamble byte.
    # STOPS - Defines the preamble bytes after which a stop bit is
    #      transmitted. For example, if STOPS=0x04, a stop bit is
    #      transmitted after the 3rd preamble byte.
    # LENGTH - Length of the preamble in bytes.
    #
    # Note: If there is a one in the same position for both STARTS and STOPS,
    #       the stop takes precedence.

    # The preamble is the device address, byte address, and (if a read) device address again:
    #   preamble[0] = 0xA0; // devAddr (write)
    #   preamble[1] = 0x00; // byteAddress (MSB)
    #   preamble[2] = 0x00; // byteAddress (LSB)
    #   preamble[3] = 0xA1; // devAddr (read)

    # from the HDL

    def i2c_configure(self, preamble_length, starts, stops, preamble):
        """Configure the buffer for the next transmission."""

        if preamble_length > 7:
            print('Preamble data is too long')
            # throw DataTooLongException();

        self.i2c['m_pBuf'] = [None]*(4+preamble_length)
        self.i2c['m_pBuf'][0] = preamble_length
        self.i2c['m_pBuf'][1] = starts
        self.i2c['m_pBuf'][2] = stops
        # Payload length will be provided later.
        self.i2c['m_pBuf'][3] = 0
        for i in range(preamble_length):
            self.i2c['m_pBuf'][4+i] = preamble[i]

        self.i2c['m_nDataStart'] = 4 + preamble_length

    def i2c_transmit(self, data, data_length):
        """Send data along the SCL and SDA lines."""

        self.i2c['m_pBuf'][3] = data_length
        for i in range(data_length):
            self.i2c['m_pBuf'].append(data[i])

        # Reset the memory pointer and transfer the buffer.
        self.fpga.xem.ActivateTriggerIn(
            self.endpoints['MEMSTART'].address, self.endpoints['MEMSTART'].bit_index_low)
        for i in range(data_length + self.i2c['m_nDataStart']):
            # print('(transmit) WireIn Value = {}'.format(self.i2c['m_pBuf'][i]))
            mask = 0xff << self.endpoints['IN'].bit_index_low
            value = self.i2c['m_pBuf'][i] << self.endpoints['IN'].bit_index_low
            self.fpga.xem.SetWireInValue(
                self.endpoints['IN'].address, value, mask)
            self.fpga.xem.UpdateWireIns()
            self.fpga.xem.ActivateTriggerIn(
                self.endpoints['MEMWRITE'].address, self.endpoints['MEMWRITE'].bit_index_low)

        # Start I2C transaction
        self.fpga.xem.ActivateTriggerIn(
            self.endpoints['START'].address, self.endpoints['START'].bit_index_low)

        # Wait for transaction to finish
        for i in range(int(I2CController.I2C_MAX_TIMEOUT_MS)):
            self.fpga.xem.UpdateTriggerOuts()
            # change to waiting for True
            if self.fpga.xem.IsTriggered(self.endpoints['DONE'].address, (1 << self.endpoints['DONE'].bit_index_low)):
                return True
            time.sleep(0.001)

        print('Timeout error in transmit')

    def i2c_receive(self, data_length, data_transfer='wire'):
        """Take in data from the SCL and SDA lines.
            
            Parameters
            ----------
            data_length : int
                Number of bytes expected to receive.
            data_transfer : str
                The form of the data transfer. Either 'wire' and 'pipe'. Defaults to 'wire'.

            Returns
            -------
            data or buf, e : list or bytearray, int
                The data or a bytearray and error code, depending on whether data_transfer was 'wire' or 'pipe'.
            """


        if data_transfer.lower() == 'pipe':
            try:
                self.endpoints['FIFO_RESET']
                self.endpoints['PIPE_OUT']
            except KeyError as e:
                raise KeyError('i2c_receive requires the I2C endpoints FIFO_RESET and PIPE_OUT. One or both are missing.')

            self.fpga.xem.ActivateTriggerIn(self.endpoints['FIFO_RESET'].address, self.endpoints['FIFO_RESET'].bit_index_low)

        self.i2c['m_pBuf'][0] |= 0x80
        self.i2c['m_pBuf'][3] = data_length

        # Reset the memory pointer and transfer the buffer.
        self.fpga.xem.ActivateTriggerIn(
            self.endpoints['MEMSTART'].address, self.endpoints['MEMSTART'].bit_index_low)

        for i in range(self.i2c['m_nDataStart']):
            # print('WireIn Value = {}'.format(self.i2c['m_pBuf'][i]))
            mask = 0xff << self.endpoints['IN'].bit_index_low
            value = self.i2c['m_pBuf'][i] << self.endpoints['IN'].bit_index_low
            self.fpga.xem.SetWireInValue(
                self.endpoints['IN'].address, value, mask)
            self.fpga.xem.UpdateWireIns()
            self.fpga.xem.ActivateTriggerIn(
                self.endpoints['MEMWRITE'].address, self.endpoints['MEMWRITE'].bit_index_low)

        # Start I2C transaction
        self.fpga.xem.ActivateTriggerIn(
            self.endpoints['START'].address,
            self.endpoints['START'].bit_index_low)

        # Wait for transaction to finish
        for _ in range(int(I2CController.I2C_MAX_TIMEOUT_MS / 10)):
            self.fpga.xem.UpdateTriggerOuts()

            if self.fpga.xem.IsTriggered(self.endpoints['DONE'].address,
                                         (1 << self.endpoints['DONE'].bit_index_low)):
                if data_transfer.lower() == 'wire':
                    # Read data: Reset the memory pointer
                    self.fpga.xem.ActivateTriggerIn(
                        self.endpoints['MEMSTART'].address, self.endpoints['MEMSTART'].bit_index_low)
                    data = [None]*data_length
                    for i in range(data_length):  # for each byte we have three API calls
                        self.fpga.xem.UpdateWireOuts()
                        data_tmp = self.fpga.xem.GetWireOutValue(
                            self.endpoints['OUT'].address)
                        mask = 0xff << self.endpoints['OUT'].bit_index_low
                        data[i] = (
                            data_tmp & mask) >> self.endpoints['OUT'].bit_index_low
                        self.fpga.xem.ActivateTriggerIn(
                            self.endpoints['MEMREAD'].address, self.endpoints['MEMREAD'].bit_index_low)
                    return data
                if data_transfer.lower() == 'pipe':
                    return self.fpga.read_pipe_out(self.endpoints['PIPE_OUT'].address, data_length)
            time.sleep(0.01)

        print('Timeout Exception in Rx')

    # def i2c_write8(self, devAddr, regAddr, data_length, data):

    #     preamble = [devAddr & 0xfe, regAddr]
    #     self.i2c_configure(2, 0x00, 0x00, preamble)
    #     return self.i2c_transmit(data, data_length)

    def i2c_write_long(self, devAddr, regAddr, data_length, data):
        """Send a write command with given data to regAddr on devAddr.

        regAddr must be given in a list."""
        if (regAddr == [None]) or (regAddr == None):
            # for chips without register addresses -- just a single register
            preamble = [devAddr & 0xfe]
        else:
            preamble = [devAddr & 0xfe] + regAddr  # + data
        # def i2c_configure(self, data_length, starts, stops, preamble)
        self.i2c_configure(len(preamble), 0x00, 1 << len(preamble), preamble)
        return self.i2c_transmit(data, data_length)

    # Sequence is
    # [START] DEV_ADDR(W) REG_ADDR [START] DEV_ADDR(R) VALUE
    def i2c_read_long(self, devAddr, regAddr, data_length, data_transfer='wire'):
        """Read data_length bytes from regAddr on devAddr.

        Parameters
        ----------
        devAddr : int
            8 bit address (don't set the read bit (LSB) since this is done in this function)
        regAddr : int
            Written to device (this is a list and must be even if length 1)
        data_length : int
            Number of bytes expected to receive
        data_transfer : str
            The form of the data transfer. Either 'wire' and 'pipe'. Defaults to 'wire'.

        Returns
        -------
        data or buf, e : list or bytearray, int
            The data or a bytearray and error code, depending on whether data_transfer was 'wire' or 'pipe'.
        """
        
        if (regAddr == None) or (regAddr == [None]):
            # for chips without register addresses -- just a single register
            preamble = [devAddr | 0x01]
            # self.i2c_configure(1, 0x01, 0x00, preamble)
            self.i2c_configure(1, 0x00, 0x00, preamble)  # no starts needed

        else:
            preamble = [devAddr & 0xfe] + regAddr + [devAddr | 0x01]
            # signature: i2c_configure(data_length, starts (a one for each byte that gets a start), stops, preamble):
            start_positions = 0x01 << len(regAddr)
            self.i2c_configure(len(preamble), start_positions, 0x00, preamble)
        data = self.i2c_receive(data_length, data_transfer)

        return data

    def reset_device(self):
        """Reset the I2C controller using an OK TriggerIn."""

        return self.fpga.xem.ActivateTriggerIn(
            self.endpoints['RESET'].address,
            self.endpoints['RESET'].bit_index_low)



from ..core import Endpoint
from ..utils import gen_mask
import copy


class SPIFifoDriven():
    """Class for SPI controllers on the FPGA driven by a FIFO.

    Attributes
    ----------
    default_data_mux : dict
        Class attribute. Default data_mux dictionary.
    fpga : FPGA
        FPGA instance this controller uses to communicate.
    endpoints : dict
        Endpoints on the FPGA this controller uses to communicate.
    master_config : int
        Value of the CTRL register in the Wishbone.
    data_mux: dict
        Name to select value dictionary for source data MUX.
    current_data_mux : str
        Name of the current MUX data source.
    """

    default_data_mux = {
        'DDR': 0,
        'host': 1,
        'ads8686_chA': 2,
        'DDR_norepeat': 3,  # does not send an update if the DDR data is the same. Reduces digital switching noise on the SPI bus
        'ad7961_ch0': 4,
        'ad7961_ch1': 5,
        'ad7961_ch2': 6,
        'ad7961_ch3': 7,
    }

    def __init__(self, fpga, endpoints, master_config, data_mux=default_data_mux):
        self.fpga = fpga
        self.endpoints = endpoints
        self.master_config = master_config
        self.data_mux = data_mux
        # Start with host to minimize unintentional commands from other sources on startup
        self.current_data_mux = None
        if self.set_data_mux('host') == -1:
            # 'host' was not in data_mux dict
            print("WARNING: no 'host' in data_mux dict")

    @classmethod
    def create_chips(cls, fpga, number_of_chips, endpoints=None, master_config=None):
        """Instantiate a number of new chips.

        The number must be an integer greater than zero. The endpoints between
        each instance will be incremented. If the endpoints argument is left
        as None, then we will use copies of the endpoints_from_defines
        dictionary for the endpoints for each instance, and update that
        original dictionary when we increment the endpoints. This way, the
        endpoints there are ready for another instantiation if needed.
        """

        if type(number_of_chips) is not int or number_of_chips <= 0:
            print('number_of_chips must be an integer greater than 0')
            return False

        chips = []
        if master_config is None:
            # Use class default for master_config
            for i in range(number_of_chips):
                chips.append(cls(fpga=fpga, endpoints=endpoints))
                # Use deepcopy to keep the endpoints for different instances separate
                endpoints = copy.deepcopy(chips[-1].endpoints)
                Endpoint.advance_endpoints(endpoints)
        else:
            for i in range(number_of_chips):
                chips.append(cls(fpga=fpga, endpoints=copy.deepcopy(
                    endpoints), master_config=master_config))
                # Use deepcopy to keep the endpoints for different instances separate
                endpoints = copy.deepcopy(chips[-1].endpoints)
                Endpoint.advance_endpoints(endpoints)

        if endpoints is None:
            # Increment shared endpoints dictionary
            # TODO: is there a better way to do this?
            # We need to get the shared endpoints in endpoints_from_defines to
            # increment and we only have the endpoints given to us in the
            # argument.
            shared_full_eps = Endpoint.endpoints_from_defines
            shared_chip_eps = shared_full_eps[
                list(shared_full_eps.keys())[
                    list(shared_full_eps.values()).index(chips[0].endpoints)
                ]
            ]
            Endpoint.advance_endpoints(shared_chip_eps, number_of_chips)
        else:
            # Increment custom dictionary
            Endpoint.advance_endpoints(endpoints, number_of_chips)

        return chips

    def filter_select(self, operation='set'):
        """Set whether SPI data comes from filter ('set') or direct ('clear')

        Direct comes from the spi_fifo_driven data.
        """

        if operation == 'set':
            self.fpga.set_wire_bit(self.endpoints['FILTER_SEL'].address,
                                   self.endpoints['FILTER_SEL'].bit_index_low)
        elif operation == 'clear':
            self.fpga.clear_wire_bit(self.endpoints['FILTER_SEL'].address,
                                     self.endpoints['FILTER_SEL'].bit_index_low)
        else:
            print(f'Incorrect operation: {operation} for filter select \n')

    def filter_sum(self, operation='set'):
        """Set whether Cmd data is summed with filter output"""

        if operation == 'set':
            self.fpga.set_wire_bit(self.endpoints['SUMMATION_ENABLE'].address,
                                   self.endpoints['SUMMATION_ENABLE'].bit_index_low)
        elif operation == 'clear':
            self.fpga.clear_wire_bit(self.endpoints['SUMMATION_ENABLE'].address,
                                     self.endpoints['SUMMATION_ENABLE'].bit_index_low)
        else:
            print(f'Incorrect operation: {operation} for filter sum \n')

    def filter_downsample(self, operation='set'):
        """Set whether filter output data is downsampled"""
        
        if operation == 'set':
            self.fpga.set_wire_bit(self.endpoints['DOWNSAMPLE_ENABLE'].address,
                                   self.endpoints['DOWNSAMPLE_ENABLE'].bit_index_low)
        elif operation == 'clear':
            self.fpga.clear_wire_bit(self.endpoints['DOWNSAMPLE_ENABLE'].address,
                                     self.endpoints['DOWNSAMPLE_ENABLE'].bit_index_low)
        else:
            print(f'Incorrect operation: {operation} for filter downsample \n')

    def set_data_mux(self, source, filter_data=False):
        """Configure the MUX that routes data source to the SPI output.

        Parameters
        ----------
        source : str
            See SPIFifoDriven.data_mux dict for options and conversion.
        """

        select_val = self.data_mux.get(source)
        if select_val is None:
            print(f'Set data mux failed, {source} not available')
            return -1

        if filter_data:
            endpoint = 'FILTER_DATA_SEL'
        else:
            endpoint = 'DATA_SEL'

        mask = gen_mask(range(self.endpoints[endpoint].bit_index_low,
                              self.endpoints[endpoint].bit_index_high))
        data = (self.data_mux[source]
                << self.endpoints[endpoint].bit_index_low)
        self.fpga.set_wire(self.endpoints[endpoint].address, data,
                           mask=mask)
        self.current_data_mux = source

    def set_clk_divider(self, divide_value):
        """Set clock divider to configure rate of SPI updates.

        Tupdate = Tclk * divide_value
        """
        mask = gen_mask(range(self.endpoints['PERIOD_ENABLE'].bit_index_low,
                              self.endpoints['PERIOD_ENABLE'].bit_index_high))

        self.fpga.set_wire(self.endpoints['PERIOD_ENABLE'].address,
                           divide_value << self.endpoints['PERIOD_ENABLE'].bit_index_low,
                           mask)

        # resets the SPI state machine
        self.fpga.xem.ActivateTriggerIn(self.endpoints['REG_TRIG'].address,
                                        self.endpoints['REG_TRIG'].bit_index_low)

    def set_ctrl_reg(self, reg_value):
        """Configures the SPI Wishbone control register over the registerBridge.

        HDL default is ctrlValue = 16'h3010 (initialized in the HDL)
        """

        self.fpga.xem.WriteRegister(
            self.endpoints['REGBRIDGE_OFFSET'].address, reg_value)

        # resets the SPI state machine -- needed since these registers are only
        #   programmed at startup of the state machine
        self.fpga.xem.ActivateTriggerIn(self.endpoints['REG_TRIG'].address,
                                        self.endpoints['REG_TRIG'].bit_index_low)

    def set_spi_sclk_divide(self, divide_value=0x01):
        """Configures the SPI Wishbone clock divider register over the registerBridge.

        HDL default is 8'h13 (initialized in the HDL)
        """

        sys_clk = 200  # in MHz
        print('SCLK predicted frequency {:.2f} [MHz]'.format(
            sys_clk/(divide_value+1)))
        self.fpga.xem.WriteRegister(
            self.endpoints['REGBRIDGE_OFFSET'].address + 1, divide_value)

        # resets the SPI state machine -- needed since these WishBone
        #   registers are only programmed at startup of the state machine
        self.fpga.xem.ActivateTriggerIn(self.endpoints['REG_TRIG'].address,
                                        self.endpoints['REG_TRIG'].bit_index_low)

    def write(self, data):
        """Host write 24 bits of data to the chip over SPI."""

        if self.current_data_mux != 'host':
            self.set_data_mux('host')
        self.fpga.set_wire(self.endpoints['HOST_WIRE_IN'].address, data)
        self.fpga.xem.ActivateTriggerIn(self.endpoints['HOST_TRIG'].address,
                                        self.endpoints['HOST_TRIG'].bit_index_low)



from ..core import Endpoint, Register
from ..utils import custom_signed_to_int
from .SPIController import SPIController
from .ADCDATA import ADCDATA

class ADS8686(SPIController, ADCDATA):
    # TODO: add class docstring

    registers = Register.get_chip_registers('ADS8686')
    msg_w = 0x8000
    range = {10:  0b00,
             2.5: 0b01,
             5:   0b10}
    lpf_khz = {39:  0b00,
               15:  0b01,
               376: 0b10}

    def __init__(self, fpga, master_config=0x3010, endpoints=None):
        # master_config=0x3010 Sets CHAR_LEN=16, ASS (auto SlaveSelect),
        # IE (interrupt enable), Tx/Rx Pos. Edge  TODO: why was this changed??
        if endpoints is None:
            endpoints = Endpoint.get_chip_endpoints('ADS8686')
        super().__init__(fpga=fpga, master_config=master_config, endpoints=endpoints)
        self.ranges = [5]*16  # voltage ranges of all 16 channels
        self.num_bits = 16
        self.name = 'ADS8686'

    def write(self, msg, reg_name):
        """Write to an internal register on the chip."""

        reg = self.registers[reg_name].address << 9
        return super().write(reg | ADS8686.msg_w | msg)

    def read(self, reg_name):
        """Read from an internal register on the chip."""

        reg = self.registers[reg_name].address << 9
        super().write(reg)
        return super().read()

    def read_channel(self, channel):
        """Read a desired channel on the chip."""

        pass  # TODO: write method

    def read_last(self):
        """Return the last value read by the ADS8686.

        This reads a wire_out rather than a pipe_out so we get 1 data point
        rather than many.
        """

        data = self.fpga.read_wire(self.endpoints['OUT'].address)
        buf = int.to_bytes(data, 4, byteorder='little')
        return self.deswizzle(buf)

    def setup(self):
        """Perform basic setup for default chip use."""

        # TODO -- defaults to modify the SPI setup
        self.set_host_mode()  # required for the SPI configuration to work
        self.configure_master_bin(self.master_config)  # configures directly
        # self.set_frequency(5)
        self.set_divider(3)  # TODO: check on scope: 200/(3+1)=50 MHz
        self.select_slave(1)

    def set_range(self, vals):
        """Set the voltage range of all channels.

        If vals is a single value then all channels are given that value.
        If vals is a list of length 16 then each channel gets a separate range
        vals must be 10, 5, or 2.5. These represent +/-10, +/-5, and +/-2.5 V.
        """

        if not isinstance(vals, list):
            vals = [vals]*16
        if len(vals) != 16:
            print('Value must be a single number of list of length 16')
        self.ranges = vals

        towrite = 0
        regs = ['rangeA1', 'rangeA2', 'rangeB1', 'rangeB2']
        for idx, val in enumerate(vals):
            towrite += ADS8686.range[val] << ((idx % 4)*2)
            if idx % 4 == 3:
                if self.fpga.debug:
                    print('Writing 0x{:x} to reg {}'.format(
                        towrite, regs[idx//4]))
                self.write(towrite, regs[idx//4])
                towrite = 0

    def reg_to_voltage(self, reg_val, chan_num=0):
        """Calculates the channel's voltage from the register value.

        Uses the stored gain range. reg_val can be list of ints or int
        """
        #  TODO: setup for sequence of channels
        val = custom_signed_to_int(reg_val, 16)  # 16-bit channel readings
        lsb = (self.ranges[chan_num]*2)/2**16
        return val*lsb

    def write_reg_bridge(self, clk_div=1000):
        """Set clk divider and spi_controller to continuously read ADC data."""

        # Configures the clock divider to determine the CONVST frequency
        self.fpga.xem.WriteRegister(self.endpoints['REGBRIDGE_OFFSET'].address + 0x0,
                                    clk_div)  # (1/200e6)*1000 = 5 us period
        # now load the sequence of wishbone commands that will be sent to the SPI converter
        self.fpga.xem.WriteRegister(self.endpoints['REGBRIDGE_OFFSET'].address + 0x1,
                                    0x8000_0001)  # Tx data register

        self.fpga.xem.WriteRegister(self.endpoints['REGBRIDGE_OFFSET'].address + 0x2,
                                    0x4000_0000)  # 0x0000 loaded into the Tx register -- for NOP

        self.fpga.xem.WriteRegister(self.endpoints['REGBRIDGE_OFFSET'].address + 0x3,
                                    0x8000_0041)  #

        self.fpga.xem.WriteRegister(self.endpoints['REGBRIDGE_OFFSET'].address + 0x4,
                                    0x4000_3110)  #
        # reset the clk divider
        self.fpga.send_trig(self.endpoints['CLK_DIV_RESET'])
        self.set_fpga_mode()  # lower the control bit for host vs. FPGA driven
        # so the FPGA takes control of sending SPI

    def set_lpf(self, lpf):
        # TODO: add method docstring

        if lpf not in ADS8686.lpf_khz.keys():
            print('Error in ADS8686 LPF setup.')
            print('   Skipping ADS8686 LPF setup')
            print('   Frequency options are: {}'.format(
                list(ADS8686.lpf_khz.keys())))
            print('-'*40)
        else:
            self.write(ADS8686.lpf_khz[lpf], 'lpf')  # 376 khZ

    def setup_sequencer(self, chan_list=[('FIXED', 'FIXED'), ('0', 'FIXED')], voltage_range=5, lpf=376):
        """Start the sequencer looping through the given channels.

        Parameters
        ----------
        chan_list : list
            Ordered list of channel pairs (A, B) to loop through. Options below.
                [0-7]
                AVDD
                ALDO
                FIXED (0xAAAA, 0x5555)
        voltage_range : int
            Voltage range of the channels, positive and negative.
        lpf : int
            Low pass frequency in kHz.
        """
        named_chans = {
            'AVDD': 0b1000,
            'ALDO': 0b1001,
            'FIXED': 0b1011,
            '0': 0,
            '1': 1,
            '2': 2,
            '3': 3,
            '4': 4,
            '5': 5,
            '6': 6,
            '7': 7,
        }
        codes = []
        for chan in chan_list:
            code = 0
            for i in range(len(chan)):
                code_piece = named_chans.get(chan[i])
                if code_piece is None:
                    print(f'ERROR: name not recognized: {chan}')
                    code = named_chans.get('FIXED') + \
                        (named_chans.get('FIXED') << 4)
                    break
                else:
                    # B side shifted 4 bits left, A side at 0
                    code += code_piece << (4 * i)
            codes.append(code)

        # Room for 32 codes, any more after that are not included
        if len(codes) > 32:
            print(
                f'WARNING: too many channels. Using first 32/{len(codes)} channels')
            codes = codes[:32]
        base_creg = 0x0000  # default value
        self.write(base_creg, 'config')
        # set channels to be the fixed digital code
        # channel B = 0x5555; channel A = 0xAAAA
        #  this is over-written by sequencer anyways
        # TODO: if this is over-written anyways, do we need it?
        self.write((0xb | (0xb << 4)), 'chan_sel')
        # set the range of all channels to +/-5V
        self.set_range(voltage_range)  # sets all channels to +/-5V
        self.set_lpf(lpf)
        # setup the sequencer
        # sequence0: fixed values and then continue
        #   8 SSREN; 7-4 CHSEL_B; 3-0 CHSEL_A
        backto_first_stack = 0x100  # SSREN is bit 8

        # Write all but the last code, which is done separately
        for i in range(len(codes) - 1):
            code = codes[i]
            self.write(code, 'seq' + str(i))
        # Write last code with loop back: use (len(codes) - 1) since i is not set if len(codes)==1
        self.write((codes[-1] | backto_first_stack),
                   'seq' + str(len(codes) - 1))
        # enable the sequencer
        self.write(base_creg | 0x20, 'config')
        return codes

    def hw_reset(self, val=True):
        """Trigger ADS8686 active low hardware reset."""

        if val:
            self.fpga.clear_wire_bit(self.endpoints['RESET'].address,
                                     self.endpoints['RESET'].bit_index_low)

        elif not val:
            self.fpga.set_wire_bit(self.endpoints['RESET'].address,
                                   self.endpoints['RESET'].bit_index_low)



from ..core import Register
from ..utils import int_to_list
from .I2CController import I2CController
import time


class TMF8801(I2CController):
    """Class for the Time-of-flight I2C device

    Subclass of the I2CController class. Attributes and methods below are
    differences in this class from I2CController only.

    Attributes
    ----------
    ADDRESS : int
        7-bit device address with R/W bit space 
    registers : dict
        Name-Register pairs for the internal registers of the chip.
    """

    ADDRESS = 0b0100_0001 << 1
    registers = Register.get_chip_registers('TMF8801')
    apps = {'measure': 0xC0, 'bootloader': 0x80}
    MULTI_BYTE_ORDER = 'LSB_1st'

    def write(self, data, register_name):
        """Write data to any register on the chip.
            first reads to enable writing of (just) bit-fields within the register
        """

        dev_addr = self.ADDRESS
        register = self.registers[register_name]
        data <<= register.bit_index_low

        # Mask only the bits for the specified register.
        mask = 0
        for bit in range(register.bit_index_high, register.bit_index_low - 1, -1):
            mask |= 0x1 << bit

        # Compare with current data to mask unchanged values
        read_out = self.read(register_name)
        if read_out == None:
            print('Read for masking FAILED')
            return False
        new_data = (data & mask) | (read_out & ~mask)

        # Turn data into a list of bytes for i2c_write method
        list_data = int_to_list(new_data)
        number_of_bytes = len(list_data)
        list_data_str = ''.join(" ".join(f'0x{x:02x}' for x in list_data))
        print(
            f'i2c write long: 0x{dev_addr:02x}, reg addr 0x{register.address:02x}, data {list_data_str}')
        self.i2c_write_long(
            dev_addr, [register.address], number_of_bytes, list_data)

    def read(self, register_name, number_of_bytes=None):
        """Return data from any register on the chip."""

        dev_addr = self.ADDRESS
        register = self.registers[register_name]
        # This is the total number of bytes to read
        byte_number = (register.bit_index_high // 8) + 1
        # e.g. 16-bit register = 2 bytes, i2c_read_long starts at the MSB so we read 2 bytes to get Byte 0
        if number_of_bytes is None:
            number_of_bytes = byte_number
        # print(f'Read slave address: 0x{dev_addr:02x}, reg addr 0x{register.address:02x}')
        read_back_list = self.i2c_read_long(
            dev_addr, [register.address], number_of_bytes)

        # Turn the list into an integer
        read_back_data = 0
        if read_back_list == None:
            return None

        if self.MULTI_BYTE_ORDER == 'MSB_1st':
            for byte in read_back_list:
                # print('Readback byte of 0x{:02X}'.format(byte))
                read_back_data <<= 8
                read_back_data |= byte
        elif self.MULTI_BYTE_ORDER == 'LSB_1st':
            for byte in read_back_list[::-1]:  # flip list order
                # print('Readback byte of 0x{:02X}'.format(byte))
                read_back_data <<= 8
                read_back_data |= byte

        # Get only the bits for the specified register from what was read back.
        desired_bits = 0
        for bit in range(register.bit_index_high, register.bit_index_low - 1, -1):
            desired_bits += 0x1 << bit
        desired_data = (read_back_data
                        & desired_bits) >> register.bit_index_low

        return desired_data

    def power_down_high_impedance(self):
        """Power down the DAC output to high impedance (default)."""

        pass
        # self.write(0b10, 'DAC_PDN')

    def get_id(self):
        """Return the device ID and rev ID as one integer."""

        device_id = self.read('ID')
        version_id = self.read('REVID')
        return (device_id << self.registers['ID'].bit_index_low) | (version_id << self.registers['ID'].bit_index_low)

    def cpu_reset(self):
        self.write(1, 'CPU_RESET')

    def cpu_ready(self):
        return self.read('CPU_READY', 1)

    def load_app(self, app='measure'):

        if app == 'measure':
            val = 0xC0
        elif app == 'bootloader':
            val = 0x80
        self.write(val, 'APPREQID')

        return self.read('APPID')

    def read_app(self):

        appid = self.read('APPID')
        appname = list(self.apps.keys())[list(self.apps.values()).index(appid)]
        print(f'In app {appname}')
        return appid

    def rom_fw_version(self):
        rev_major = self.read('APPREV_MAJOR', number_of_bytes=1)
        rev_minor = self.read('APPREV_MINOR', number_of_bytes=1)
        patch = self.read('APPREV_PATCH', number_of_bytes=1)

        print(f'major: {rev_major}, minor: {rev_minor}, patch: {patch}')

        return rev_major, rev_minor, patch

    def bl_command(self, cmd):

        appid = self.read('APPID')
        if appid == self.apps['bootloader']:
            print('cannot send boatloader command when in measure app')
            return 0
        if cmd == 'ramremap_reset':
            val = 0x11
        elif cmd == 'download_init':
            val = 0x14

        self.write(val, 'CMD_DATA7')

    def ram_write_status(self):
        # TODO: Host_Driver_Comm document suggests this reads back 3 bytes
        return self.read('CMD_DATA7', number_of_bytes=3)

    def download_init(self):

        dev_addr = self.ADDRESS
        reg_addr = self.registers['CMD_DATA7'].address
        # TODO: why 0x29 - specified by host driver comms
        data = [0x14, 0x01, 0x29, 0xC1]
        self.i2c_write_long(dev_addr, [reg_addr],
                            len(data), data)

    def ramremap_reset(self):

        dev_addr = self.ADDRESS
        reg_addr = self.registers['CMD_DATA7'].address
        data = [0x11, 0x00, 0xEE]
        self.i2c_write_long(dev_addr, [reg_addr],
                            len(data), data)

    def read_dist_peak(self):

        dev_addr = self.ADDRESS
        reg_addr = self.registers['DISTANCE_PEAK'].address
        num_bytes = self.registers['DISTANCE_PEAK'].bit_width//8
        return self.i2c_read_long(dev_addr, [reg_addr],
                                  data_length=num_bytes)

    def read_data(self):
        """ For correctly updating of these registers by TMF8801, an I2C block read starting from address 0x1D
        until 0x27 shall be done.
        0x1D, 0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27"""

        dev_addr = self.ADDRESS
        reg_addr = self.registers['STATUS'].address

        read_data = self.i2c_read_long(dev_addr, [reg_addr],
                                       data_length=11)

        vals = {'status': read_data[0],
                'register_contents': read_data[1],
                'transaction_id': read_data[2],
                'result_num': read_data[3],
                'result_info': read_data[4],
                'dist_mm': read_data[5] + (read_data[6] << 8),
                'sys_clk': read_data[7] + (read_data[8] << 8) + (read_data[9] << 16) + (read_data[10] << 24)}
        vals['sys_clk_seconds'] = vals['sys_clk']*0.2e-6
        return vals, read_data

    def read_by_addr(self, reg_addr, num_bytes=1):
        dev_addr = self.ADDRESS
        return self.i2c_read_long(dev_addr, [reg_addr],
                                  data_length=num_bytes)

    def factory_calibration(self, filename=None):
        """
        Perform factory calibration and return 14 bytes of data
        See AN000597 -- Factor Calibration 8.1

        Calibration Environment: Device has to be in the final (correct) optical stack
        Clear glass (no smudge on the glass)
        No target in front of the device within 40 cm (see datasheet)
        Dark room or low ambient light
        """
        print('Factory calibration: device should be in dark room with no objects within 40 cm')

        self.write(0x0A, 'COMMAND')  # Command is register 0x10
        time.sleep(2)
        cal_done = 0
        count = 0
        while (cal_done != 0x0A) and (count < 20):
            cal_done = self.read('REGISTER_CONTENTS')
            time.sleep(0.1)
            count = count + 1
            print(f'Cal done 0x{cal_done:02x}')

        cal_data = self.read_by_addr(0x20, num_bytes=14)
        print('Calibration data')
        for d in cal_data:
            print(d)

        with open(filename, 'w') as f:
            f.write(','.join(cal_data))

        return cal_data



from ..core import Endpoint
from .SPIFifoDriven import SPIFifoDriven
import numpy as np


class AD5453(SPIFifoDriven):
    # TODO: add class docstring

    bits = 12,
    vref = 2.5*2

    def __init__(self, fpga, master_config=0x3010, endpoints=None, data_mux=SPIFifoDriven.default_data_mux):

        if endpoints is None:
            endpoints = Endpoint.get_chip_endpoints('AD5453')

        # master_config=0x3010 Sets CHAR_LEN=16, ASS, IE
        super().__init__(fpga=fpga, master_config=master_config,
                         endpoints=endpoints, data_mux=data_mux)
        # Default to clocking data into the shift register on the falling edge of the clock
        self.clk_edge_bits = 0b00

        # TODO: cleanup! 500kHz is within set_filter_coeff
        self.filter_coeff = {0: 0x009e1586,
                             1: 0x20000000,
                             2: 0x40000000,
                             3: 0x20000000,
                             4: 0xbce3be9a,
                             5: 0x12f3f6b0,
                             8: 0x7fffffff,
                             9: 0x20000000,
                             10: 0x40000000,
                             11: 0x20000000,
                             12: 0xab762783,
                             13: 0x287ecada,
                             7: 0x7fffffff,
                             15: 0x0000_2000}  # offset=0, scale=1

        self.filter_offset = 4
        self.filter_len = np.max(
            list(self.filter_coeff.keys())) - np.min(list(self.filter_coeff.keys()))
        self.current_data_mux = None
        self.set_data_mux('host')

    def set_clk_rising_edge(self):
        """Set the signals we write to use the rising edge of the clock.

        Default is falling edge. To return to the falling edge, the chip
        requires a power cycle (turn it off and back on).
        """

        self.clk_edge_bits = 0b11

    def set_ctrl_reg(self, reg_value=0x3010):
        """Configures the SPI Wishbone control register over the registerBridge.

        reg_value=0x3010 sets CHAR_LEN=16, ASS, IE
        """

        super().set_ctrl_reg(reg_value=reg_value)

    def write_filter_coeffs(self):
        # TODO: add method docstring

        # TODO is this correct? and how to parameterize?
        for i in np.arange(self.filter_offset, 1 + self.filter_offset + self.filter_len):
            if (i-self.filter_offset) in self.filter_coeff:
                addr = int(
                    self.endpoints['REGBRIDGE_OFFSET'].address + i)
                val = int(self.filter_coeff[i-self.filter_offset])
                # TODO: ian has first addr of 0x19, second as 0x1a
                print(
                    'Write filter addr=0x{:02X}  value=0x{:02X}'.format(addr, val))
                self.fpga.xem.WriteRegister(addr, val)

    def write_filter_coeffs_simultaneous(self):
        # TODO: add method docstring

        # https://opalkelly.com/examples/setting-and-getting-multiple-registers/#tab-python
        loop_thru = np.arange(self.filter_offset, 1
                              + self.filter_offset + self.filter_len)

        # ok should already be imported from Endpoint import
        regs = ok.okTRegisterEntries((len(loop_thru)))

        for i in loop_thru:  # TODO is this correct? and how to parameterize?
            if (i-self.filter_offset) in self.filter_coeff:
                addr = int(
                    self.endpoints['REGBRIDGE_OFFSET'].address + i)
                val = int(self.filter_coeff[i-self.filter_offset])
                idx = int(i-self.filter_offset)
                regs[idx].address = addr
                regs[idx].data = val
                # TODO: ian has first addr of 0x19, second as 0x1a
                print(
                    'Write filter addr=0x{:02X}  value=0x{:02X}'.format(addr, val))
        self.fpga.xem.WriteRegisters(regs)

    def read_coeff_debug(self):
        """
        Read two wire outs for coefficients a3 section1, scale 3
        debug wires available for filters 0, 1
        """

        debug_coeff_0 = self.fpga.read_wire(
            self.endpoints[f'COEFF_DEBUG_0'].address)
        debug_coeff_1 = self.fpga.read_wire(
            self.endpoints[f'COEFF_DEBUG_1'].address)
        debug_coeff_2 = self.fpga.read_wire(
            self.endpoints[f'COEFF_DEBUG_2'].address)
        debug_coeff_3 = self.fpga.read_wire(
            self.endpoints[f'COEFF_DEBUG_3'].address)

        print(f'Read Coefficients:')
        print('    0:', debug_coeff_0)
        print('    1:', debug_coeff_1)
        print('    2:', debug_coeff_2)
        print('    3:', debug_coeff_3)
        return debug_coeff_0, debug_coeff_1, debug_coeff_2, debug_coeff_3

    def change_filter_coeff(self, target, value=None):
        # TODO: add method docstring

        if (target == 'passthru') or (target == 'passthrough'):
            self.filter_coeff = {#0: 0x7fff_ffff,   # changed for about unity gain
                                 0: 0x80000000,
                                 1: 0x20000000,
                                 2: 0,
                                 3: 0,
                                 4: 0,
                                 5: 0,
                                 8: 0x7fffffff,
                                 9: 0x20000000,
                                 10: 0,
                                 11: 0,
                                 12: 0,
                                 13: 0,
                                 7: 0x7fffffff,
                                 6: 0,
                                 15: 0x0000_0900} # offset=0, scale=1 
        elif target == '100kHz':
            self.filter_coeff = {0: 0x0000_6f84,
                                 1: 0x20000000,
                                 2: 0x40000000,
                                 3: 0x20000000,
                                 4: 0x8e301ca0,
                                 5: 0x32b7759a,
                                 8: 0x7fffffff,
                                 9: 0x20000000,
                                 10: 0x40000000,
                                 11: 0x20000000,
                                 12: 0x86d2475f,
                                 13: 0x3a2447ec,
                                 7: 0x7fffffff, 
                                 15: 0x0000_1000} # offset=0, scale=1 
        elif target == '500kHz':
            self.filter_coeff = {#0: 0x009e1586,
                                 0: 0xFF61EA7A,
                                 1: 0x20000000,
                                 2: 0x40000000,
                                 3: 0x20000000,
                                 4: 0xbce3be9a,
                                 5: 0x12f3f6b0,
                                 8: 0x7fffffff,
                                 9: 0x20000000,
                                 10: 0x40000000,
                                 11: 0x20000000,
                                 12: 0xab762783,
                                 13: 0x287ecada,
                                 7: 0x7fffffff,
                                 15: 0x0000_0900}
                                 #15: 0x0000_0B1F} # offset=0, scale=1
        else :
            self.filter_coeff = value



from ..core import Register
from ..utils import int_to_list, from_voltage
from .I2CController import I2CController

class DAC53401(I2CController):
    """Class for the I2C DAC chip DAC53401.

    Subclass of the I2CController class. Attributes and methods below are
    differences in this class from I2CController only.

    Attributes
    ----------
    ADDRESS_HEADER : int
        7-bit device address with R/W bit space and 4 MSBs filled in specific
        to this chip. The rest is left as 0 to be filled in later.
    registers : dict
        Name-Register pairs for the internal registers of the TCA9555 chip.
    addr_pins : int
        3 LSBs of the 7-bit device address formed alongside the address header
        used to differentiate between different instances of the TCA9555 chip.
    """

    ADDRESS_HEADER = 0b10010000
    registers = Register.get_chip_registers('DAC53401')

    #        Address Pins Guide
    # Slave Address   |   A0 Pin
    #           000   |   AGND
    #           001   |   VDD
    #           010   |   SDA
    #           011   |   SCL

    def write(self, data, register_name='DAC_DATA'):
        """Write data to any register on the chip."""

        dev_addr = DAC53401.ADDRESS_HEADER | (self.addr_pins << 1)
        register = DAC53401.registers[register_name]
        data <<= register.bit_index_low

        # Mask only the bits for the specified register.
        mask = 0
        for bit in range(register.bit_index_high, register.bit_index_low - 1, -1):
            mask |= 0x1 << bit

        # Compare with current data to mask unchanged values
        read_out = self.read(register_name)
        if read_out == None:
            print('Read for masking FAILED')
            return False
        new_data = (data & mask) | (read_out & ~mask)

        # Turn data into a list of bytes for i2c_write method
        # Essentially round up to the closest byte
        #number_of_bytes = (register.bit_width + 7) // 8
        list_data = int_to_list(new_data)
        number_of_bytes = len(list_data)
        self.i2c_write_long(
            dev_addr, [register.address], number_of_bytes, list_data)

    def read(self, register_name='DAC_DATA'):
        """Return data from any register on the chip."""

        dev_addr = DAC53401.ADDRESS_HEADER | (self.addr_pins << 1)
        register = DAC53401.registers[register_name]
        # Ex. 16-bit register: | Byte 1 [15:8] | Byte 0 [7:0] |
        byte_number = register.bit_index_high // 8
        # 16-bit register = 2 bytes, i2c_read_long starts at the MSB so we read 2 bytes to get Byte 0
        number_of_bytes = 2 - byte_number
        read_back_list = self.i2c_read_long(
            dev_addr, [register.address], number_of_bytes)

        # Turn the list into an integer
        read_back_data = 0
        if read_back_list == None:
            return None
        # First byte in the list is the MSB, shift and append the next byte
        for byte in read_back_list:
            print('Readback byte of {:02X}'.format(byte))
            read_back_data <<= 8
            read_back_data |= byte
        # Get only the bits for the specified register from what was read back.
        desired_bits = 0
        for bit in range(register.bit_index_high, register.bit_index_low - 1, -1):
            desired_bits += 0x1 << bit
        desired_data = (read_back_data
                        & desired_bits) >> register.bit_index_low

        return desired_data

    def write_voltage(self, voltage):
        """Write a voltage output from the DAC."""

        voltage_data = from_voltage(
            voltage=voltage, num_bits=10, voltage_range=5, with_negatives=False)

        self.write(voltage_data, 'DAC_DATA')

    def enable_internal_reference(self):
        """Enable the DAC's internal reference.

        This is necessary to set the gain value.
        """

        self.write(0x1, 'REF_EN')

    def set_gain(self, gain):
        """Set the DAC's output gain value.

        Gain can be set to 1.5x, 2x, 3x, or 4x the internal reference.
        Internal reference must be enabled.
        """

        gain_dict = {
            1.5: 0b00,
            2: 0b01,
            3: 0b10,
            4: 0b11
        }

        gain_code = gain_dict.get(gain)
        if gain_code == None:
            print('Invalid gain')
            return False
        self.write(gain_code, 'DAC_SPAN')

    def get_gain(self):
        """Return the DAC's current output gain value."""

        gain_dict = {
            0b00: 1.5,
            0b01: 2,
            0b10: 3,
            0b11: 4
        }

        gain_code = self.read('DAC_SPAN')
        return gain_dict.get(gain_code)

    def config_func(self, function_name):
        """Configure the function generator.

        'triangle': Triangle wave between MARGIN_HIGH code to MARGIN_LOW code
            with slope defined by SLEW_RATE.
        'sawtooth_falling': Saw-Tooth wave between MARGIN_HIGH code to
            MARGIN_LOW code with slope defined by SLEW_RATE and immeidate falling edge.
        'sawtooth_rising': Saw-Tooth wave between MARGIN_HIGH code to
            MARGIN_LOW code with slope defined by SLEW_RATE and immeidate rising edge.
        'square': Square wave between MARGIN_HIGH code to MARGIN_LOW code with
            pulse high and low period defined by defined by SLEW_RATE.
        """

        func_dict = {
            'triangle': 0b00,
            'sawtooth_falling': 0b01,
            'sawtooth_rising': 0b10,
            'square': 0b11
        }

        func_code = func_dict.get(function_name.lower())
        if func_code == None:
            print('Invalid function name')
            return False
        self.write(func_code, 'FUNC_CONFIG')

    def start_func(self):
        """Start function generation.

        Waveform configured through config_func() and config_margins() functions.
        """
        self.write(0b1, 'START_FUNC_GEN')

    def stop_func(self):
        """Stop function generation."""

        self.write(0b0, 'START_FUNC_GEN')

    def config_margins(self, margin_high=None, margin_low=None):
        """Configure margin high and low values."""

        if margin_high != None:
            self.write(margin_high, 'MARGIN_HIGH')

        if margin_low != None:
            self.write(margin_low, 'MARGIN_LOW')

    def config_step(self, step):
        """Configure the number of bits to step through."""

        step_dict = {
            1: 0b000,
            2: 0b001,
            3: 0b010,
            4: 0b011,
            6: 0b100,
            8: 0b101,
            16: 0b110,
            32: 0b111
        }

        step_code = step_dict.get(step)
        if step_code == None:
            print('Invalid step')
            return False
        self.write(step_code, 'CODE_STEP')

    def config_rate(self, rate):
        """Configure the rate to step through each bit."""

        # Instead of asking the user to type in the timing, the rates are organized slowest to fastest and numbered 1-16
        rate_dict = {
            1: 0b1011,  # 1638.4 * 1.75 us
            2: 0b1010,  # 1638.4 * 1.50 us
            3: 0b1001,  # 1638.4 * 1.25 us
            4: 0b1000,  # 1638.4        us
            5: 0b0111,  # 204.8  * 1.75 us
            6: 0b0110,  # 204.8  * 1.50 us
            7: 0b0101,  # 204.8  * 1.25 us
            8: 0b0100,  # 204.8         us
            9: 0b0011,  # 25.6   * 1.75 us
            10: 0b0010,  # 25.6   * 1.50 us
            11: 0b0001,  # 25.6   * 1.25 us
            12: 0b0000,  # 25.6          us
            13: 0b1100,  # 12            us
            14: 0b1101,  # 8             us
            15: 0b1110,  # 4             us
            16: 0b1111  # None
        }

        rate_code = rate_dict.get(rate)
        if rate_code == None:
            print('Invalid rate')
            return False
        self.write(rate_code, 'SLEW_RATE')

    def reset(self):
        """Reset the chip using a software reset."""

        self.write(0b1010, 'SW_RESET')

    def lock(self):
        """Lock the registers of the device so they cannot be changed."""

        self.write(0b1, 'DEVICE_LOCK')

    def unlock(self):
        """Unlock the registers of the device so they can be changed again."""

        self.write(0b0101, 'DEVICE_UNLOCK_CODE')

    def power_up(self):
        """Power up the DAC output."""

        self.write(0b00, 'DAC_PDN')

    def power_down_10k(self):
        """Power down the DAC output to 10K ohms."""

        self.write(0b01, 'DAC_PDN')

    def power_down_high_impedance(self):
        """Power down the DAC output to high impedance (default)."""

        self.write(0b10, 'DAC_PDN')

    def get_id(self):
        """Return the device ID and version ID as one integer."""

        device_id = self.read('DEVICE_ID')
        version_id = self.read('VERSION_ID')
        return (device_id << DAC53401.registers['DEVICE_ID'].bit_index_low) | (version_id << DAC53401.registers['VERSION_ID'].bit_index_low)



from ..core import Register
from .I2CController import I2CController


class DAC101C081(I2CController):
    """Class for the DAC101C081.

    Subclass of the I2CController class. Attributes and methods below are
    differences in this class from I2CController only.

    Attributes
    ----------
    ADDRESS_HEADER : int
        7-bit device address with R/W bit space.
    registers : dict
        Name-Register pairs (this chip has only 1 register! PD[13:12], DATA[9:0])
        PD = b00 for normal operation  
    addr_pins : int
        3 LSBs of the 7-bit device address formed alongside the address header
        used to differentiate between different instances using the ADDR pins.
    """

    ADDRESS_HEADER = (0b000_1101 << 1)  # ADR0=Gnd -- Bath Clamp. ADR0 to GND
    # The device address table in the data sheet is strange
    # see here:
    #   https://e2e.ti.com/support/data-converters-group/data-converters/f/data-converters-forum/955893/dac101c081-dac101c081-i2c-address-selection
    registers = Register.get_chip_registers('DAC101C081')
    addr_pins = 0
    dac_res_bits = 10

    def write(self, data):
        """Write DAC data and set the PD bits to 00
           DAC data should be no greater than 2^{dac_res_bits} - 1
        """
        dev_addr = self.ADDRESS_HEADER | (self.addr_pins << 1)
        if self.dac_res_bits == 10:
            data = data << 2  # for the 10 bit DAC the 2 LSBs are zeros.
        elif self.dac_res_bits == 12:
            pass  # don't need to shift the data
        else:
            raise ValueError('Incorrect DAC resolution in bits')
        # set PD bits to b00 [13:12] also bits 15:14 are do not care
        data = data & 0x0fff
        # Turn data into a list of bytes for i2c_write method
        # equivalent to [((data & 0xff00)>>8), (data & 0x00ff)]
        list_data = [data // (16**2), data % (16**2)]
        self.i2c_write_long(dev_addr, [None], 2, list_data)

    def read(self):
        """Read 2 bytes of data from the chip register"""

        dev_addr = self.ADDRESS_HEADER | (self.addr_pins << 1) | 0b1
        data_list = self.i2c_read_long(dev_addr, [None], 2)
        data = ((data_list[0] & 0x0f) << 6) + (data_list[1] >> 2)
        pd = ((data_list[0] & 0x30) >> 4)
        return data, pd

    def power_down(self):
        data = 0b0010_0000_0000_0000  # this is the 100 kOhm power down mode
        dev_addr = self.ADDRESS_HEADER | (self.addr_pins << 1)

        # Turn data into a list of bytes for i2c_write method
        # equivalent to [((data & 0xff00)>>8), (data & 0x00ff)]
        list_data = [data // (16**2), data % (16**2)]
        self.i2c_write_long(dev_addr, [None], 2, list_data)
